import pytest


from labrad.decorators import setting
from labrad.server import LabradServer
from labrad import types


class DummyServer(LabradServer):
    @setting(100, param0='s: this is a string',
                  param1='b: this is a boolean')
    def dummy_setting(self, c, param0, param1):
        pass


def test_setting_types_with_comments():
    accepted_tags = DummyServer.dummy_setting.accepts
    accepted_types = [types.parseTypeTag(s) for s in accepted_tags]
    assert len(accepted_types) == 1
    assert accepted_types[0] == types.parseTypeTag('sb')


if __name__ == '__main__':
    pytest.main(['-v', __file__])



# Copyright (C) 2007  Matthew Neeley
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
labrad.client

Contains a blocking client connection to labrad.
"""

import warnings

from labrad import constants as C, types as T
from labrad.backend import ManagerService
from labrad.concurrent import map_future
from labrad.errors import Error
from labrad.support import (mangle, indent, PrettyMultiDict, FlatPacket,
                            PacketRecord, PacketResponse, hexdump)

class NotFoundError(Error):
    code = 10
    def __init__(self, key):
        self.msg = 'Could not find "%s".' % key

def unwrap(s, after='|'):
    def trim(line):
        return line.split(after, 1)[-1]
    return '\n'.join(trim(line) for line in s.split('\n'))

class SettingWrapper(object):
    """Object to wrap a single setting on a single server.

    Information about the setting is loaded on demand and cached.
    You can force a refresh of this information by calling the refresh()
    method. Calling this object directly will send a request and block until
    the result is available. Calling the .future method will send a request and
    return a future with which to get the result later.

    The keyword argument unflatten=False will suppress the normal unflattening
    and return a FlatData object.
    """
    def __init__(self, server, name, pyName, ID):
        self.name = self.__name__ = self._labrad_name = name
        self.ID = ID
        self._py_name = pyName
        self._server = server
        self._mgr = server._mgr
        self._refreshed = False

    def __call__(self, *args, **kw):
        wait = kw.pop('wait', True)
        if not wait:
            warnings.warn("Calling settings with wait=False is deprecated. "
                          "Use setting.future(...) instead.")
        f = self.future(*args, **kw)
        return f.result() if wait else f

    def future(self, *args, **kw):
        wrap = kw.pop('wrap', True)
        tag = kw.pop('tag', None) or self.accepts
        if not len(args):
            args = None
        elif len(args) == 1:
            args = args[0]
        flat = T.flatten(args, tag)
        f = self._server._send([(self.ID, flat)], **kw)
        if wrap:
            f = map_future(f, lambda resp: resp[0][1])
        return f

    # data to be loaded on demand
    @property
    def accepts(self):
        self._refresh()
        return self._accepts

    @property
    def returns(self):
        self._refresh()
        return self._returns

    @property
    def description(self):
        self._refresh()
        return self.__doc__

    @property
    def notes(self):
        self._refresh()
        return self._notes

    def _refresh(self):
        if not self._refreshed:
            info = self._mgr.getSettingInfoByName(self._server.ID, self.name)
            self.__doc__, self._accepts, self._returns, self._notes, self.ID = info
            self._refreshed = True

    def refresh(self):
        # mark that we need to refresh.  This doesn't happen immediately,
        # but instead only when needed later due to property access
        self._refreshed = False

    def __repr__(self):
        return unwrap("""\
            |LabRAD Setting: "%s" >> "%s" (ID=%d)
            |
            |%s
            |
            |Accepts:
            |%s
            |
            |Returns:
            |%s
            |
            |%s """) % (self._server.name, self.name, self.ID, self.description,
                        indent('\n'.join(self.accepts)),
                        indent('\n'.join(self.returns)), self.notes)


class DynamicAttrDict(PrettyMultiDict):
    _parent = None

    def __getitem__(self, key):
        try:
            return super(DynamicAttrDict, self).__getitem__(key)
        except KeyError:
            # force refresh and try again
            if self._parent:
                self._parent.refresh(now=True)
            try:
                return super(DynamicAttrDict, self).__getitem__(key)
            except KeyError:
                raise NotFoundError(key)


class HasDynamicAttrs(object):
    """An object with attributes looked up dynamically on labrad."""
    def __init__(self):
        self.__attrs = None
        self.__cache = None
        self._refreshed = False

    @property
    def _attrs(self):
        if not self._refreshed:
            self._refresh()
        return self.__attrs

    def _fixName(self, name):
        pyName = mangle(name)
        if pyName in self._staticAttrs:
            pyName = 'lr_' + pyName
        return pyName

    def _refresh(self):
        """Update the list of available attributes."""
        try:
            if self._refreshed:
                return

            if self.__attrs is None:
                self.__attrs = DynamicAttrDict()
                self.__attrs._parent = self
                self.__cache = {}

            # get current list of attributes
            attrs = self._getAttrs()
            if len(attrs):
                names, IDs = zip(*attrs)
            else:
                names, IDs = [], []
            pyNames = [self._fixName(name) for name in names]
            attrs = zip(names, pyNames, IDs)

            # delete names of old attributes (but leave them in the cache)
            deletions = [pyName for pyName in self.__attrs if pyName not in pyNames]
            for pyName in deletions:
                del self.__attrs[pyName]

            # add new attributes
            additions = [(n, p, ID) for (n, p, ID) in attrs
                                    if p not in self.__attrs]
            for name, pyName, ID in additions:
                if name in self.__cache:
                    # pull from cache if possible
                    s = self.__cache[name]
                    s.ID = ID # update attribute ID
                else:
                    s = self._wrapAttr(self, name, pyName, ID)
                self.__cache[name] = s
                self.__attrs[pyName, name, ID] = s

            # refresh all attributes
            for attr in self.__attrs.values():
                if hasattr(attr, 'refresh'):
                    attr.refresh()

            self._refreshed = True
        except Exception, e:
            print 'Error refreshing dynamic attributes:'
            print e, repr(e)

    def _getAttrs(self):
        """Get the current list of attributes from labrad.

        Should be overridden by subclasses with a method
        to get the attributes they need.
        """
        return []

    _staticAttrs = [] # static attributes names, so dynamic names don't collide

    def _wrapAttr(self, *args):
        """Should be overridden by subclasses"""

    def refresh(self, now=False):
        """Signal that a refresh is needed.

        If now is true, the refresh happens immediately,
        otherwise it is put off until actually needed later.
        """
        self._refreshed = False
        if now:
            self._refresh()

    def __dir__(self):
        """Return a list of attributes for tab-completion.
        """
        self.refresh() # force refresh so the list is current
        return sorted(set(self._attrs.keys() + self.__dict__.keys() + dir(type(self))))
    
    def __getattr__(self, key):
        return self._attrs[key]

    def __getitem__(self, key):
        return self._attrs[key]


class ServerWrapper(HasDynamicAttrs):
    """A wrapper for a labrad server."""

    def __init__(self, cxn, name, pyName, ID, context=None):
        HasDynamicAttrs.__init__(self)
        self._cxn = cxn
        self._mgr = cxn._mgr
        self.name = self._labrad_name = name
        self._py_name = pyName
        self.ID = ID
        self._ctx = context

    _staticAttrs = ['settings', 'context', 'packet', 'sendMessage']
    _wrapAttr = SettingWrapper

    def _getAttrs(self):
        info = self._mgr.getServerInfo(self.ID)
        doc, notes, self._slist = info
        self.__doc__ = doc + '\n' + notes
        return self._slist

    @property
    def settings(self):
        self._refresh()
        return self._attrs

    def context(self):
        return self._cxn.context()

    def packet(self, **kw):
        return PacketWrapper(self, **kw)

    def __call__(self, context=None):
        """Create a new server wrapper based on this one but with a new default context."""
        if context is None:
            context = self._cxn.context()
        return ServerWrapper(self._cxn, self.name, self._py_name, self.ID, context=context)

    def _send(self, *args, **kw):
        if 'context' not in kw or kw['context'] is None:
            kw['context'] = self._ctx
        return self._cxn._send(self.ID, *args, **kw)

    def sendMessage(self, ID, *args, **kw):
        """Send a message to this server."""
        if 'context' not in kw or kw['context'] is None:
            kw['context'] = self._ctx
        tag = kw.pop('tag', [])
        if len(args) == 0:
            args = None
        elif len(args) == 1:
            args = args[0]
        self._cxn._sendMessage(self.ID, [(ID, args, tag)], **kw)

    def __repr__(self):
        self._refresh()
        return unwrap("""\
            |LabRAD Server: %s (ID=%d)
            |
            |%s
            |
            |Settings:
            |%s""") % (self.name, self.ID, self.__doc__,
                       indent(repr(self.settings)))


class PacketWrapper(HasDynamicAttrs):
    """An object to encapsulate a labrad packet to a server."""

    def __init__(self, server, **kw):
        HasDynamicAttrs.__init__(self)
        self._server = server
        self._packet = []
        self._kw = kw

    def send(self, wait=True, **kw):
        """Send this packet to the server and wait for the result.

        Using the keyword argument unflatten=False will suppress the
        default unflattening and return FlatData objects"""
        if not wait:
            warnings.warn("Sending packets with wait=False is deprecated. "
                          "Use packet.send_future(...) instead.")
        f = self.send_future(**kw)
        f.result()
        return f.result() if wait else f

    def send_future(self, **kw):
        """Send this packet to the server and get the result as a future."""
        records = [(rec.ID, rec.flat) for rec in self._packet]
        f = self._server._send(records, **dict(self._kw, **kw))
        f = map_future(f, PacketResponse, self._server, self._packet)
        return f

    @property
    def settings(self):
        self._refresh()
        return self._attrs

    def to_cluster(self, context=None):
        """Get a representation of this packet as a flattenable cluster.

        This method returns a representation of this packet as a cluster of
        (context, name, records), where records is itself a cluster of
        (setting name, value), one for each setting record in the packet.
        """
        if context is None:
            context = self._kw.get('context', self._server._ctx)
            if context is None:
                context = (self._server._cxn.ID, 0)
        records = tuple((rec.name, rec.flat) for rec in self._packet)
        return FlatPacket(context, self._server.name, records)

    _staticAttrs = ['settings', 'send', 'send_future', 'to_cluster']

    def _getAttrs(self):
        """Grab the list of the server's attributes."""
        self._server._refresh() # ensure refresh
        return self._server._slist

    def _wrapAttr(self, _parent, name, pyName, ID):
        s = self._server.settings[name]
        def wrapped(*args, **kw):
            key = kw.pop('key', None)
            tag = kw.pop('tag', None) or s.accepts
            if not len(args):
                args = None
            elif len(args) == 1:
                args = args[0]
            flat = T.flatten(args, tag)
            rec = PacketRecord(ID=s.ID, data=args, tag=tag, flat=flat, key=key,
                               name=name)
            self._packet.append(rec)
            return self
        return wrapped

    def __setitem__(self, key, value):
        """Update existing parts of the packet, indexed by key."""
        for i, rec in enumerate(self._packet):
            if key == rec.key:
                flat = T.flatten(value, rec.tag)
                self._packet[i] = rec._replace(data=value, flat=flat)

    def _dataStr(self, data):
        if len(data) < 160:
            return self._dataRepr(data)
        else:
            x = self._dataRepr(data[0:32])
            '\n'.join('    ' + y for y in x.splitlines())
            return "string of length %d beginning with... \n%s" % (len(data), x)

    def _dataRepr(self, data):
        if all((ord(x) > 31 and ord(x) < 127) or x.isspace() for x in data): # Is string printable ascii
            return data
        else:
            return hexdump(data)

    def _recordRepr(self, rec, short=False):
        """Create a string representation of a packet record.

        Includes the setting name and result key (if specified), as well as
        a possibly truncated repr of the data for this setting.
        """
        key_str = "" if rec.key is None else " (key=%s)" % (rec.key,)
        prefix = '%s%s: ' % (rec.name, key_str)
        data_str = self._dataStr(str(rec.data)) if short else self._dataRepr(str(rec.data))
        data_str = data_str.replace('\n', '\n' + ' '*len(prefix))
        return prefix + data_str

    def __str__(self):
        return self.__repr__(True)

    def __repr__(self, short=False):
        """Create a string representation of this packet.

        Includes the server the packet is intended for, as well as all the
        records that have been added to the packet.
        """
        data_str = '\n'.join(self._recordRepr(rec, short=short) for rec in self._packet)
        return unwrap("""\
            |Packet for server: '%s'
            |
            |Data:
            |%s""") % (self._server.name, indent(data_str))


class Client(HasDynamicAttrs):
    def __init__(self, cxn, context=None):
        HasDynamicAttrs.__init__(self)
        self._backend = cxn
        self._mgr = ManagerService(cxn)
        if context is None:
            context = cxn.context()
        self._ctx = context

    def __enter__(self):
        """Enter the body of a with statement."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit the body of a with statement."""
        try:
            self.disconnect()
        except:
            pass
        return False

    def __call__(self, context=None):
        return Client(self._backend, context)

    def _getAttrs(self):
        if not self.connected:
            return []
        return self._mgr.getServersList()

    _staticAttrs = ['servers', 'connect', 'disconnect', 'context', 'spawn']
    _wrapAttr = ServerWrapper

    @property
    def servers(self):
        self._refresh()
        return self._attrs

    # attributes proxied to the underlying connection, which may be shared among multiple clients
    @property
    def name(self):
        return self._backend.name

    @property
    def ID(self):
        return self._backend.ID

    @property
    def host(self):
        return self._backend.host

    @property
    def port(self):
        return self._backend.port

    @property
    def connected(self):
        return self._backend.connected

    def connect(self, host, port=None, timeout=C.TIMEOUT, password=None,
                tls_mode=C.MANAGER_TLS):
        self._backend.connect(host, port=port, timeout=timeout,
                              password=password, tls_mode=tls_mode)

    def disconnect(self):
        self._backend.disconnect()

    def context(self):
        return self._backend.context()

    def spawn(self):
        """Start a new independent connection to the same manager."""
        return Client(self._backend.spawn())

    def _send(self, target, records, *args, **kw):
        """Send a packet over this connection."""
        if 'context' not in kw or kw['context'] is None:
            kw['context'] = self._ctx
        return self._backend.sendRequest(target, records, *args, **kw)

    def _sendMessage(self, target, records, *args, **kw):
        """Send a message over this connection."""
        if 'context' not in kw or kw['context'] is None:
            kw['context'] = self._ctx
        return self._backend.sendMessage(target, records, *args, **kw)

    def __repr__(self):
        if self.connected:
            return unwrap("""\
                |LabRAD Client: '%s' on %s:%s
                |
                |Available servers:
                |%s""") % (self.name, self.host, self.port,
                           indent(repr(self.servers)))
        else:
            return unwrap("""\
                |LabRAD Client: '%s'
                |
                |Disconnected""") % self.name





# Copyright (c) 2001-2008 Twisted Matrix Laboratories.
# See LICENSE for details.


"""
This module provides support for Twisted to be driven by the Qt mainloop.

In order to use this support, simply do the following::
    |  app = QApplication(sys.argv) # your code to init Qt
    |  import qt4reactor
    |  qt4reactor.install()

alternatively:

    |  from twisted.application import reactors
    |  reactors.installReactor('qt4')

Then use twisted.internet APIs as usual.  The other methods here are not
intended to be called directly.

If you don't instantiate a QApplication or QCoreApplication prior to
installing the reactor, a QCoreApplication will be constructed
by the reactor.  QCoreApplication does not require a GUI so trial testing
can occur normally.

Twisted can be initialized after QApplication.exec_() with a call to
reactor.runReturn().  calling reactor.stop() will unhook twisted but
leave your Qt application running

API Stability: stable

Maintainer: U{Glenn H Tarbox, PhD<mailto:glenn@tarbox.org>}

Previous maintainer: U{Itamar Shtull-Trauring<mailto:twisted@itamarst.org>}
Original port to QT4: U{Gabe Rudy<mailto:rudy@goldenhelix.com>}
Subsequent port by therve
"""

__all__ = ['install']


import sys, time

from zope.interface import implements

from PyQt4.QtCore import QSocketNotifier, QObject, SIGNAL, QTimer, QCoreApplication
from PyQt4.QtCore import QEventLoop

from twisted.internet.interfaces import IReactorFDSet
from twisted.python import log
from twisted.internet.posixbase import PosixReactorBase

class TwistedSocketNotifier(QSocketNotifier):
    """
    Connection between an fd event and reader/writer callbacks.
    """

    def __init__(self, reactor, watcher, type):
        QSocketNotifier.__init__(self, watcher.fileno(), type)
        self.reactor = reactor
        self.watcher = watcher
        self.fn = None
        if type == QSocketNotifier.Read:
            self.fn = self.read
        elif type == QSocketNotifier.Write:
            self.fn = self.write
        QObject.connect(self, SIGNAL("activated(int)"), self.fn)


    def shutdown(self):
        QObject.disconnect(self, SIGNAL("activated(int)"), self.fn)
        self.setEnabled(False)
        self.fn = self.watcher = None
        self.deleteLater()


    def read(self, sock):
        w = self.watcher
        #self.setEnabled(False)    # ??? do I need this?
        def _read():
            why = None
            try:
                why = w.doRead()
            except:
                log.err()
                why = sys.exc_info()[1]
            if why:
                self.reactor._disconnectSelectable(w, why, True)
            elif self.watcher:
                pass
                #self.setEnabled(True)
        log.callWithLogger(w, _read)
        self.reactor.reactorInvocation()

    def write(self, sock):
        w = self.watcher
        self.setEnabled(False)
        def _write():
            why = None
            try:
                why = w.doWrite()
            except:
                log.err()
                why = sys.exc_info()[1]
            if why:
                self.reactor._disconnectSelectable(w, why, False)
            elif self.watcher:
                self.setEnabled(True)
        log.callWithLogger(w, _write)
        self.reactor.reactorInvocation()

class fakeApplication(QEventLoop):
    def __init__(self):
        QEventLoop.__init__(self)

    def exec_(self):
        QEventLoop.exec_(self)

class QTReactor(PosixReactorBase):
    """
    Qt based reactor.
    """
    implements(IReactorFDSet)

    _timer = None

    def __init__(self):
        self._reads = {}
        self._writes = {}
        self._timer = QTimer()
        self._timer.setSingleShot(True)
        if QCoreApplication.startingUp():
            self.qApp = QCoreApplication([])
            self._ownApp = True
        else:
            self.qApp = QCoreApplication.instance()
            self._ownApp = False
        self._blockApp = None
        self._readWriteQ = []

        """ some debugging instrumentation """
        self._doSomethingCount = 0

        PosixReactorBase.__init__(self)

    def addReader(self, reader):
        if not reader in self._reads:
            self._reads[reader] = TwistedSocketNotifier(self, reader,
                                                       QSocketNotifier.Read)


    def addWriter(self, writer):
        if not writer in self._writes:
            self._writes[writer] = TwistedSocketNotifier(self, writer,
                                                        QSocketNotifier.Write)


    def removeReader(self, reader):
        if reader in self._reads:
            #self._reads[reader].shutdown()
            #del self._reads[reader]
            self._reads.pop(reader).shutdown()

    def removeWriter(self, writer):
        if writer in self._writes:
            self._writes[writer].shutdown()
            #del self._writes[writer]
            self._writes.pop(writer)


    def removeAll(self):
        return self._removeAll(self._reads, self._writes)


    def getReaders(self):
        return self._reads.keys()


    def getWriters(self):
        return self._writes.keys()

    def callLater(self, howlong, *args, **kargs):
        rval = super(QTReactor, self).callLater(howlong, *args, **kargs)
        self.reactorInvocation()
        return rval

    def crash(self):
        super(QTReactor, self).crash()

    def iterate(self, delay=0.0):
        t = self.running # not sure I entirely get the state of running
        self.running = True
        self._timer.stop() # in case its not (rare?)
        try:
            if delay == 0.0:
                self.reactorInvokePrivate()
                self._timer.stop() # supports multiple invocations
            else:
                endTime = delay + time.time()
                self.reactorInvokePrivate()
                while True:
                    t = endTime - time.time()
                    if t <= 0.0: return
                    self.qApp.processEvents(QEventLoop.AllEvents |
                                      QEventLoop.WaitForMoreEvents, t*1010)
        finally:
            self.running = t

    def addReadWrite(self, t):
        self._readWriteQ.append(t)

    def runReturn(self, installSignalHandlers=True):
        QObject.connect(self._timer, SIGNAL("timeout()"),
                        self.reactorInvokePrivate)
        self.startRunning(installSignalHandlers=installSignalHandlers)
        self._timer.start(0)

    def run(self, installSignalHandlers=True):
        try:
            if self._ownApp:
                self._blockApp = self.qApp
            else:
                self._blockApp = fakeApplication()
            self.runReturn(installSignalHandlers)
            self._blockApp.exec_()
        finally:
            self._timer.stop() # should already be stopped

    def reactorInvocation(self):
        self._timer.setInterval(0)

    def reactorInvokePrivate(self):
        if not self.running:
            self._blockApp.quit()
        self._doSomethingCount += 1
        self.runUntilCurrent()
        t = self.timeout()
        if t is None:
            t = 0.1
        else:
            t = min(t, 0.1)
        self._timer.setInterval(t*1010)
        self.qApp.processEvents() # could change interval
        self._timer.start()

    def doIteration(self):
        assert False, "doiteration is invalid call"

def install():
    """
    Configure the twisted mainloop to be run inside the qt mainloop.
    """
    from twisted.internet import main
    reactor = QTReactor()
    main.installReactor(reactor)



# Copyright (C) 2007  Matthew Neeley
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
labrad.util.hydrant

Generate random LabRAD data for use in testing.
"""

from random import choice, randint, gauss
from datetime import datetime, timedelta

from labrad import types as T

def randType(noneOkay=True, listOkay=True, nStructs=0):
    choices = [
        T.LRNone,
        T.LRBool,
        T.LRInt,
        T.LRWord,
        T.LRStr,
        T.LRBytes,
        T.LRTime,
        lambda: T.LRValue(randUnits()),
        lambda: T.LRComplex(randUnits()),
        lambda: T.LRCluster(*[randType(noneOkay=False, nStructs=nStructs+1)
                              for _ in range(randint(1, 5))]),
        lambda: T.LRList(randType(noneOkay=False, listOkay=False,
                                  nStructs=nStructs+1),
                         depth=randint(1, 3)),
    ]
    if not noneOkay:
        choices = choices[1:]
    if nStructs >= 3:
        choices = choices[:-2]
    elif not listOkay:
        choices = choices[:-1]
    return choice(choices)()

def randUnits():
    return choice(('', 's', 'ms', 'us', 'm', 'm/s', 'V^2/Hz', 'V/Hz^1/2'))

def randValue(t):
    if isinstance(t, T.LRNone): return genNone()
    if isinstance(t, T.LRBool): return genBool()
    if isinstance(t, T.LRInt): return genInt()
    if isinstance(t, T.LRWord): return genWord()
    if isinstance(t, T.LRStr): return genStr()
    if isinstance(t, T.LRBytes): return genStr()
    if isinstance(t, T.LRTime): return genTime()
    if isinstance(t, T.LRComplex): return genComplex(t.unit) # check complex before value
    if isinstance(t, T.LRValue): return genValue(t.unit)
    if isinstance(t, T.LRCluster): return genCluster(*t.items)
    if isinstance(t, T.LRList): return genList(t.elem, t.depth)


def genNone(): return None
def genBool(): return choice((True, False))
def genInt(): return int(randint(-2**31, 2**31-1))
def genWord(): return long(randint(0, 2**32-1))
def genStr():
    return ''.join(chr(randint(0, 255)) for _ in xrange(randint(0, 100)))

def genTime():
    diff = timedelta(seconds=randint(-2**20, 2**20),
                     microseconds=randint(-2**20, 2**20))
    return datetime.now() + diff

def genValue(unit=None):
    return T.Value(gauss(0, 1), unit)

def genComplex(unit=None):
    return T.Complex(complex(gauss(0, 1), gauss(0, 1)), unit)

def genList(elem, depth=1):
    lengths = [randint(1, 2**(5-depth)) for _ in xrange(depth)]
    def genNDList(ls):
        if len(ls) == 1:
            return [randValue(elem) for _ in xrange(ls[0])]
        else:
            return [genNDList(ls[1:]) for _ in xrange(ls[0])]
    return genNDList(lengths)

def genCluster(*items):
    return tuple(randValue(t) for t in items)


def hoseDown(setting, n=1000, silent=True):
    for _ in range(n):
        t = randType()
        v = randValue(t)
        if not silent:
            print t
        try:
            resp = setting(v)
            assert v == resp
        except:
            print 'problem:', str(t), repr(t)
            print str(T.flatten(v)[1]), str(T.flatten(resp)[1])
            raise

def hoseDataVault(dv, n=1000, silent=True):
    for i in range(n):
        t = randType(noneOkay=False)
        v = randValue(t)
        if not silent:
            print t
        try:
            pname = 'p%03s' % i
            dv.add_parameter(pname, v)
            resp = dv.get_parameter(pname)
            assert v == resp
        except:
            print 'problem:', str(t), repr(t)
            print str(T.flatten(v)[1]), str(T.flatten(resp)[1])
            raise


if __name__ == '__main__':
    import labrad
    cxn = labrad.connect()
    try:
        hoseDown(cxn.python_test_server.echo, n=10000, silent=False)
    except Exception, e:
        print e
    else:
        print 'Success!'
    finally:
        print 'press <enter> to finish...'
        raw_input()



# Copyright (C) 2007  Matthew Neeley
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
labrad.errors

Exception classes to provide convenient, informative error and debugging
information.
"""

from labrad.types import Error

## Network

class ConnectionTimeoutError(Exception):
    """Attempt to connect to labrad took too long."""

class LoginFailedError(Exception):
    """Failed to log in to LabRAD manager."""

## Servers

class DeviceNotSelectedError(Error):
    """No device has been selected in this context."""
    code = 1

class NoDevicesAvailableError(Error):
    """No devices are available."""
    code = 2

class NoSuchDeviceError(Error):
    """No such device exists."""
    code = 3

class RequestTimeoutError(Error):
    """The request timed out."""
    code = 4

class RequestResponseError(Error):
    code = 5
    def __init__(self, errors):
        messages = ['error in setting %d (code=%d)\n%s' % (ID, err.code, err.message)
                    for (ID, err) in errors]
        self.msg = '\n'.join(messages)



"""
Run a node service for controlling other labrad servers.

This files makes it possible to run the node as an executable module:

`python -m labrad.node`
"""

import labrad.node

labrad.node.main()



# Copyright (C) 2007  Matthew Neeley
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime
import unittest
import numpy as np

import sys
import os
sys.path.insert(0, os.path.abspath('../..'))

import labrad.types as T
import labrad.units as U

from labrad.units import Value, ValueArray, Complex


class LabradTypesTests(unittest.TestCase):

    def testTags(self):
        """Test the parsing of type tags into LRType objects."""
        tests = {
            '_': T.LRNone(),
            'b': T.LRBool(),
            'i': T.LRInt(),
            'w': T.LRWord(),
            's': T.LRStr(),
            't': T.LRTime(),
            'y': T.LRBytes(),

            # clusters
            'ii': T.LRCluster(T.LRInt(), T.LRInt()),
            'b(t)': T.LRCluster(T.LRBool(), T.LRCluster(T.LRTime())),
            '(ss)': T.LRCluster(T.LRStr(), T.LRStr()),
            '(s)': T.LRCluster(T.LRStr()),
            '((siw))': T.LRCluster(T.LRCluster(T.LRStr(), T.LRInt(),
                                               T.LRWord())),

            # lists
            '*b': T.LRList(T.LRBool()),
            '*_': T.LRList(),
            '*2b': T.LRList(T.LRBool(), depth=2),
            '*2_': T.LRList(depth=2),
            '*2v[Hz]': T.LRList(T.LRValue('Hz'), depth=2),
            '*3v': T.LRList(T.LRValue(), depth=3),
            '*v[]': T.LRList(T.LRValue(''), depth=1),

            # unit types
            'v': T.LRValue(),
            'v[]': T.LRValue(''),
            'v[m/s]': T.LRValue('m/s'),
            'c': T.LRComplex(),
            'c[]': T.LRComplex(''),
            'c[m/s]': T.LRComplex('m/s'),

            # errors
            'E': T.LRError(),
            'Ew': T.LRError(T.LRWord()),
            'E(w)': T.LRError(T.LRCluster(T.LRWord())),

            # more complex stuff
            '*b*i': T.LRCluster(T.LRList(T.LRBool()), T.LRList(T.LRInt())),
        }
        for tag, type_ in tests.items():
            self.assertEqual(T.parseTypeTag(tag), type_)
            newtag = str(type_)
            if isinstance(type_, T.LRCluster) and tag[0] + tag[-1] != '()':
                # just added parentheses in this case
                self.assertEqual(newtag, '(%s)' % tag)
            else:
                self.assertEqual(newtag, tag)

    def testTagComments(self):
        """Test the parsing of type tags with comments and whitespace."""
        tests = {
            '': T.LRNone(),
            ' ': T.LRNone(),
            ': this is a test': T.LRNone(),
            '  : this is a test': T.LRNone(),
            '   i  ': T.LRInt(),
            '   i  :': T.LRInt(),
            '   i  : blah': T.LRInt(),
        }
        for tag, type_ in tests.items():
            self.assertEqual(T.parseTypeTag(tag), type_)

    def testDefaultFlatAndBack(self):
        """
        Test roundtrip python->LabRAD->python conversion.

        No type requirements are given in these tests. In other words, we allow
        pylabrad to choose a default type for flattening.

        In this test, we expect A == unflatten(*flatten(A)). In other words,
        we expect the default type chosen for each object to unflatten as
        an object equal to the one originally flattened.
        """
        tests = [
            # simple types
            None,
            True, False,
            1, -1, 2, -2, 0x7FFFFFFF, -0x80000000,
            1L, 2L, 3L, 4L, 0L, 0xFFFFFFFFL,
            '', 'a', '\x00\x01\x02\x03',
            datetime.now(),

            # values
            5.0,
            Value(6, ''),
            Value(7, 'ms'),
            8+0j,
            Complex(9+0j, ''),
            Complex(10+0j, 'GHz'),

            # ValueArray and ndarray
            # These types should be invariant under flattening followed by
            # unflattening. Note, however, that since eg. [1, 2, 3] will
            # unflatten as ndarray with dtype=int32, we do not put lists
            # in this test.
            U.ValueArray([1, 2, 3], 'm'),
            U.ValueArray([1j, 2j, 3j], 's'),
            np.array([1, 3, 4], dtype='int32'),
            np.array([1.1, 2.2, 3.3]),

            # clusters
            (1, True, 'a'),
            ((1, 2), ('a', False)),

            # lists
            [],
            #[1, 2, 3, 4],
            #[1L, 2L, 3L, 4L],
            [[]],
            [['a', 'bb', 'ccc'], ['dddd', 'eeeee', 'ffffff']],

            # more complex stuff
            [(1L, 'a'), (2L, 'b')],
        ]
        for data_in in tests:
            data_out = T.unflatten(*T.flatten(data_in))
            if isinstance(data_in, U.ValueArray):
                self.assertTrue(data_in.allclose(data_out))
            elif isinstance(data_in, np.ndarray):
                np.testing.assert_array_equal(data_out, data_in)
            else:
                self.assertEqual(data_in, data_out)

    def testDefaultFlatAndBackNonIdentical(self):
        """
        Test flattening/unflattening of objects which change type.

        No type requirements are given in these tests. In other words, we allow
        pylabrad to choose a default type for flattening.

        In this test, we do not expect A == unflatten(*flatten(A)). This is
        mostly because list of numbers, both with an without units, should
        unflatten to ndarray or ValueArray, rather than actual python lists.
        """
        def compareValueArrays(a, b):
            """I check near equality of two ValueArrays"""
            self.assertTrue(a.allclose(b))

        tests = [
            ([1, 2, 3], np.array([1, 2, 3], dtype='int32'),
                np.testing.assert_array_equal),
            ([1.1, 2.2, 3.3], np.array([1.1, 2.2, 3.3], dtype='float64'),
                np.testing.assert_array_almost_equal),
            (np.array([3, 4], dtype='int32'), np.array([3, 4], dtype='int32'),
                np.testing.assert_array_equal),
            (np.array([1.2, 3.4]), np.array([1.2, 3.4]),
                np.testing.assert_array_almost_equal),
            ([Value(1.0, 'm'), Value(3.0, 'm')], ValueArray([1.0, 3.0], 'm'),
                compareValueArrays),
            ([Value(1.0, 'm'), Value(10, 'cm')], ValueArray([1.0, 0.1], 'm'),
                compareValueArrays),
            (ValueArray([1, 2], 'Hz'), ValueArray([1, 2], 'Hz'),
                compareValueArrays),
            (ValueArray([1.0, 2], ''), np.array([1.0, 2]),
                np.testing.assert_array_almost_equal),
            # Numpy scalar types
            (np.bool8(True), True, self.assertEqual) 
        ]
        for input, expected, comparison_func in tests:
            unflat = T.unflatten(*T.flatten(input))
            if isinstance(unflat, np.ndarray):
                self.assertEqual(unflat.dtype, expected.dtype)
            comparison_func(unflat, expected)

    def testFlatAndBackWithTypeRequirements(self):
        tests = [
            ([1, 2, 3], ['*i'], np.array([1, 2, 3]),
                np.testing.assert_array_equal),
            ([1, 2], ['*v[]'], np.array([1, 2]),
                np.testing.assert_array_almost_equal),
            ([1.1, 2.], ['*v[]'], np.array([1.1, 2.], dtype='float64'),
                np.testing.assert_array_almost_equal)
        ]
        for input, types, expected, comparison_func in tests:
            flat = T.flatten(input, types)
            unflat = T.unflatten(*flat)
            comparison_func(expected, unflat)

    def testBooleanArrayFlattening(self):
        flat = T.flatten([True, False, True])
        unflat = T.unflatten(*flat)
        flat2 = T.flatten(unflat)
        unflat2 = T.unflatten(*flat2)
        np.testing.assert_array_equal(unflat, unflat2)

    def testFailedFlattening(self):
        """
        Trying to flatten data to an incompatible type should raise an error.
        """
        cases = [
            # Simple cases
            (1, ['s', 'v[Hz]']),
            ('X', ['i', 'v', 'w']),
            (5.0, ['s', 'b', 't', 'w', 'i', 'v[Hz]']),
            # Value
            (5.0, 'v[Hz]'),
            (Value(4, 'm'), 'v[]'),
            (Value(3, 's'), ['v[Hz]', 'i', 'w']),
            # ndarray
            (np.array([1, 2, 3], dtype='int32'), '*v[Hz]'),
            (np.array([1.0, 2.4]), ['*i', '*w']),
            # ValueArray
            (U.ValueArray([1, 2, 3], 'm'), '*v[s]'),
            (U.ValueArray([1, 2], 'm'), '*v[]')
        ]
        for data, targetTag in cases:
            self.assertRaises(T.FlatteningError, T.flatten, data, targetTag)

    def testTypeHints(self):
        """Test conversion to specified allowed types."""
        passingTests = [
            # convert to default type
            (1, [], 'i'),

            # convert to first compatible type
            (1, ['s', 'w'], 'w'),
            (1, ['s', 'v'], 'v[]'),
            (1*U.m, ['s', 'v[m]'], 'v[m]'),
            # 'v' not allowed on wire
            (3.0, 'v', 'v[]'),
            (3, 'v', 'v[]'),

            # empty list gets type from hint
            ([], ['s', '*(ww)'], '*(ww)'),

            # handle unknown pieces inside clusters and lists
            (['a', 'b'], ['*?'], '*s'),
            ((1, 2, 'a'), ['ww?'], 'wws'),
            ((1, 1L), ['??'], 'iw'),
        ]
        for data, hints, tag in passingTests:
            self.assertEqual(T.flatten(data, hints)[1], T.parseTypeTag(tag))

    def testTypeSpecialization(self):
        """Test specialization of the type during flattening."""
        tests = [
            # specialization without hints
            ([([],), ([5.0],)], '*(*v)'),
            ([([],), ([Value(5, 'm')],)], '*(*v[m])'),
        ]
        for data, tag in tests:
            self.assertEqual(T.flatten(data)[1], T.parseTypeTag(tag))

    def testUnitTypes(self):
        """Test flattening with units.

        The flattening code should not do unit conversion,
        but should leave that up to the LabRAD manager to handle.
        Basically, for purposes of flattening, a unit is a unit.
        """
        tests = [
            (Value(5.0, 'ft'), ['v[m]'], 'v[ft]'),

            # real value array
            (U.ValueArray([1, 2, 3], ''), [], '*v[]'),
            (U.ValueArray([1, 2, 3], 'm'), ['*v[m]'], '*v[m]'),

            # complex value array
            (U.ValueArray([1j, 2j, 3j], ''), [], '*c[]'),
            (U.ValueArray([1j, 2j, 3j], 'm'), [], '*c[m]')
        ]
        for data, hints, tag in tests:
            self.assertEqual(T.flatten(data, hints)[1], T.parseTypeTag(tag))

        # we disallow flattening a float to a value with units,
        # as this is a major source of bugs
        try:
            T.flatten(5.0, 'v[m]')
        except Exception:
            pass
        else:
            raise Exception('Cannot flatten float to value with units')

    def testNumpySupport(self):
        """Test flattening and unflattening of numpy arrays"""
        # TODO: flesh this out with more array types
        a = np.array([1, 2, 3, 4, 5], dtype='int32')
        b = T.unflatten(*T.flatten(a))
        self.assertTrue(np.all(a == b))
        self.assertTrue(T.flatten(np.int32(5))[0] == '\x00\x00\x00\x05')
        self.assertTrue(T.flatten(np.int64(-5))[0] == '\xff\xff\xff\xfb')
        self.assertTrue(len(T.flatten(np.float64(3.15))[0]) == 8)
        with self.assertRaises(T.FlatteningError):
            T.flatten(np.int64(-5), T.LRWord())

    def testNumpyArrayScalar(self):
        with self.assertRaises(TypeError):
            T.flatten(np.array(5))
        with self.assertRaises(TypeError):
            T.flatten(U.ValueArray(np.array(5), 'ns'))


    def testIntegerRanges(self):
        """Test flattening of out-of-range integer values"""
        tests = [
            (0x80000000, 'i'),
            (-0x80000001, 'i'),
            (0x100000000, 'w'),
            (-1, 'w')
        ]
        for n, t in tests:
            with self.assertRaises(T.FlatteningError):
                T.flatten(n, t)

    def testFlattenIsIdempotent(self):
        flat = T.flatten(0x1, 'i')
        self.assertEquals(T.flatten(flat), flat)
        self.assertEquals(T.flatten(flat, 'i'), flat)
        with self.assertRaises(T.FlatteningError):
            T.flatten(flat, 'v')

    def testEvalDatetime(self):
        data = datetime.now()
        data2 = T.evalLRData(repr(data))
        self.assertEquals(data, data2)

    def testUnicodeBytes(self):
        foo = T.flatten('foo bar')
        self.assertEquals(foo, T.flatten(u'foo bar'))
        self.assertEquals(str(foo.tag), 's')
        self.assertEquals(T.unflatten(foo.bytes, 'y'), 'foo bar')
        self.assertEquals(T.unflatten(*T.flatten('foo bar', ['y'])), 'foo bar')

    def testFlattenIntArrayToValueArray(self):
        x = np.array([1, 2, 3, 4], dtype='int64')
        flat = T.flatten(x, '*v')
        y = T.unflatten(*flat)
        self.assertTrue(np.all(x == y))

    def testFlattenArrayToClusterList(self):
        """Fail if trying to flatten a numpy array to type with incorrect shape.

        See https://github.com/labrad/pylabrad/issues/290.
        """
        arr = np.arange(5, dtype='float64')
        with self.assertRaises(T.FlatteningError):
            T.flatten(arr, types=['*(v, v)'])

    def testCanFlattenFlatData(self):
        x = ('this is a test', -42, [False, True])
        flat = T.flatten(x)
        self.assertEquals(T.parseTypeTag(flat.tag), T.parseTypeTag('si*b'))
        flat2 = T.flatten(x)
        self.assertEquals(flat2, flat)
        flat3 = T.flatten(x, 'si*b')
        self.assertEquals(flat3, flat)
        with self.assertRaises(T.FlatteningError):
            T.flatten(x, 'sv')

    def testCanFlattenListOfPartialFlatData(self):
        x1 = ('this is a test', -42, [False, True])
        piece1 = T.flatten(x1)
        x2 = ('this is also a test', -43, [False, True, True, True])
        piece2 = T.flatten(x2)

        not_flattened = [x1, x2]
        partially_flattened = [piece1, piece2]
        tag = '*(si*b)'

        expected = T.flatten(not_flattened)

        flat1 = T.flatten(partially_flattened)
        self.assertEquals(flat1, expected)

        flat2 = T.flatten(partially_flattened, tag)
        self.assertEquals(flat2, expected)

        with self.assertRaises(T.FlatteningError):
            T.flatten(partially_flattened, '*(si)')

    def testCanFlattenClusterOfPartialFlatData(self):
        x1 = ('this is a test', -42, [False, True])
        piece1 = T.flatten(x1)
        x2 = ('this is also a test', -43, [False, True, True, True])
        piece2 = T.flatten(x2)

        not_flattened = (('1', x1), ('2', x2, False))
        partially_flattened = (('1', piece1), ('2', piece2, False))
        tag = '((s(si*b)) (s(si*b)b))'

        expected = T.flatten(not_flattened)

        flat1 = T.flatten(partially_flattened)
        self.assertEquals(flat1, expected)

        flat2 = T.flatten(partially_flattened, tag)
        self.assertEquals(flat2, expected)

        with self.assertRaises(T.FlatteningError):
            T.flatten(partially_flattened, '*(s(si*b))')


if __name__ == "__main__":
    unittest.main()



# Copyright (C) 2007  Matthew Neeley
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
labrad.devices

Superclass of generic device servers.
"""

from labrad import errors
from labrad.server import LabradServer, setting
from labrad.errors import Error
from labrad.support import MultiDict

from twisted.internet import defer, reactor
from twisted.internet.defer import inlineCallbacks, returnValue

LOCK_TIMEOUT = 10

class DeviceLockedError(Error):
    """The device is locked."""
    # TODO should tell who holds lock and when it expires
    code = 4

class DeviceWrapper(object):
    """A wrapper for a device."""
    def __init__(self, guid, name):
        self.guid = guid # globally-unique identifier
        self.name = name
        self.locked = False
        self._lockContext = None
        self._unlockCall = None

    def lock(self, c, timeout_s=None):
        """Get or renew a lock on this device."""
        if not self.accessibleFrom(c):
            raise DeviceLockedError()
        self.locked = True
        self._lockContext = c
        timeout_s = timeout_s or LOCK_TIMEOUT
        if self._unlockCall is None:
            self._unlockCall = reactor.callLater(timeout_s, self.unlock)
        else:
            self._unlockCall.reset(timeout_s)

    def unlock(self, c=None):
        """Release the lock on this device."""
        c = c or self._lockContext
        if not self.accessibleFrom(c):
            raise DeviceLockedError()
        self.locked = False
        if self._unlockCall and self._unlockCall.active():
            self._unlockCall.cancel()
        self._unlockCall = None
        self._lockContext = None

    def lockedInContext(self, c):
        return self.locked and (self._lockContext == c)

    def accessibleFrom(self, c):
        return (not self.locked) or (self._lockContext == c)

    def connect(self, *args, **kw):
        """Connect to this device.

        This method will be called with the args and kw args returned
        by findDevices, when a new device is created.
        """

    def shutdown(self):
        """Called when we close the connection to this device."""

    def select(self, context):
        """Select this device in a context."""

    def deselect(self, context):
        """Deselect this device in a context."""

class DeviceServer(LabradServer):
    """A server for devices.

    Creates a DeviceWrapper for each device it finds, based on a
    user-provided function.  Provides standard settings for listing
    devices, selecting a device for the current context, and
    refreshing the list of devices.  Also, provides for device-locking
    with timeouts.
    """
    name = 'Generic Device Server'
    deviceWrapper = DeviceWrapper

    def initServer(self):
        self.devices = MultiDict() # aliases -> device
        self.device_guids = {} # name -> guid
        self._next_guid = 0
        self._refreshLock = defer.DeferredLock()
        return self.refreshDeviceList()

    def stopServer(self):
        if hasattr(self, 'devices'):
            ds = [defer.maybeDeferred(dev.shutdown)
                  for dev in self.devices.values()]
            return defer.DeferredList(ds, fireOnOneErrback=True)

    def findDevices(self):
        """Return a list of found devices.

        The result should be a list of (name, args, kw) tuples
        where args and kw are the arguments tuple and keyword dict
        that will be used to call the device's connect function.
        """
        return []

    def refreshDeviceList(self):
        """Refresh the list of available devices.

        Devices are assigned a unique identifying
        number that will persist between refreshes, so that
        clients that have selected devices in context will still
        be able to refer to them after the refresh.
        """
        return self._refreshLock.run(self._doRefresh)

    def chooseDeviceWrapper(self, *args, **kw):
        """
        Override in subclass to support multiple device wrapper classes

        args and kw come from findDevices (ie same as are passed into the
        device wrapper's connect method).
        """
        return self.deviceWrapper

    @inlineCallbacks
    def _doRefresh(self):
        """Do the actual refreshing."""
        self.log('refreshing device list...')
        all_found = yield self.findDevices()

        # If there are devices for which we don't have wrappers,
        # create them. If we have a wrapper for a
        # device that is no longer available, then that wrapper should
        # be marked somehow to indicate that it is no longer available.

        def fixFound(f):
            if isinstance(f, str):
                f = f, (), {}
            name = f[0]
            args = f[1] if len(f) > 1 else ()
            kw = f[2] if len(f) > 2 else {}
            return name, args, kw
        all_found = [fixFound(f) for f in all_found]

        additions = [(name, args, kw) for (name, args, kw) in all_found
                     if name not in self.devices]
        names_found = [name for (name, args, kw) in all_found]
        deletions = [name for name in self.device_guids
                     if name in self.devices and name not in names_found]
        self.log('all_found: %s' % all_found)
        self.log('additions: %s' % additions)
        self.log('deletions: %s' % deletions)

        # start additions
        for name, args, kw in additions:
            if name in self.device_guids:
                # we've seen this device before
                # so we'll reuse the old guid
                guid = self.device_guids[name]
            else:
                guid = self.device_guids[name] = self._next_guid
                self._next_guid += 1

            deviceWrapper = self.chooseDeviceWrapper(name, *args, **kw)
            dev = deviceWrapper(guid, name)
            yield dev.connect(*args, **kw)
            self.devices[guid, name] = dev

        # do deletions
        for name in deletions:
            # we delete the device, but not its guid, so that
            # if this device comes back, users who have
            # selected it by guid will reconnect seamlessly
            dev = self.devices[name]
            del self.devices[name]
            try:
                yield dev.shutdown()
            except Exception, e:
                self.log('Error while shutting down device "%s": %s' % (name, e))

    def serverConnected(self, ID, name):
        self.refreshDeviceList()

    def serverDisconnected(self, ID, name):
        self.refreshDeviceList()

    def expireContext(self, c):
        """Release selected/locked device when context expires."""
        if 'device' in c:
            alias = c['device']
            try:
                dev = self.devices[alias]
                if dev.lockedInContext(c):
                    dev.unlock(c)
                dev.deselect(c)
            except KeyError:
                pass

    def deviceLists(self):
        """Get parallel lists of device names and IDs."""
        guids = sorted(self.devices.keys())
        names = [self.devices[g].name for g in guids]
        return guids, names

    def selectedDevice(self, context):
        """Get the selected device from the given context."""
        if not len(self.devices):
            raise errors.NoDevicesAvailableError()
        try:
            key = context['device']
        except KeyError:
            raise errors.DeviceNotSelectedError()
        try:
            dev = self.devices[key]
        except KeyError:
            raise errors.NoSuchDeviceError()
        if not dev.accessibleFrom(context.ID):
            raise DeviceLockedError()
        return dev

    def selectDevice(self, context, key=None):
        """Select a device in our current context."""
        if not len(self.devices):
            raise errors.NoDevicesAvailableError()
        if key is None:
            # use the first device
            key = sorted(self.devices.keys())[0]
        try:
            dev = self.devices[key]
        except KeyError:
            raise errors.NoSuchDeviceError(key)
        if not dev.accessibleFrom(context.ID):
            raise DeviceLockedError()

        if 'device' in context:
            if context['device'] != dev.guid:
                try:
                    oldDev = self.devices[context['device']]
                except KeyError:
                    pass
                else:
                    # we're trying to select a new device.
                    # make sure to unlock previously selected device
                    if oldDev.lockedInContext(context.ID):
                        oldDev.unlock(context.ID)
                    oldDev.deselect(context)
                context['device'] = dev.guid
                dev.select(context)
        else:
            context['device'] = dev.guid
            dev.select(context)
        return dev

    def deselectDevice(self, context):
        if 'device' in context:
            try:
                oldDev = self.devices[context['device']]
            except KeyError:
                pass
            else:
                # unlock and deselect device
                if oldDev.lockedInContext(context.ID):
                    oldDev.unlock(context.ID)
                oldDev.deselect(context)
            del context['device']

    def getDevice(self, context, key=None):
        if not len(self.devices):
            raise errors.NoDevicesAvailableError()
        if key is None:
            # use the first device
            key = sorted(self.devices.keys())[0]
        try:
            dev = self.devices[key]
        except KeyError:
            raise errors.NoSuchDeviceError()
        if not dev.accessibleFrom(context.ID):
            raise DeviceLockedError()
        return dev

    # server settings

    @setting(1, 'List Devices', returns=['*(ws)'])
    def list_devices(self, c):
        """List available devices."""
        IDs, names = self.deviceLists()
        return zip(IDs, names)

    @setting(2, 'Select Device',
                key=[': Select first device',
                     's: Select device by name',
                     'w: Select device by ID'],
                returns=['s: Name of the selected device'])
    def select_device(self, c, key=0):
        """Select a device for the current context."""
        dev = self.selectDevice(c, key=key)
        return dev.name

    @setting(3, 'Deselect Device', returns=[''])
    def deselect_device(self, c):
        """Select a device for the current context."""
        dev = self.deselectDevice(c)

    @setting(4, 'Refresh Devices', returns=['*(ws)'])
    def refresh_devices(self, c):
        """Refresh the list of available devices."""
        yield self.refreshDeviceList()
        returnValue(self.list_devices(c))

    @setting(1000001, 'Lock Device',
                      timeout=[': Lock the selected device for default time',
                            'v[s]: Lock for specified time'],
                      returns=[''])
    def lock_device(self, c, timeout):
        """Lock a device to be accessible only in this context."""
        dev = self.selectedDevice(c)
        if timeout is not None:
            timeout = timeout['s']
        dev.lock(c.ID, timeout)

    @setting(1000002, 'Release Device', returns=[''])
    def release_device(self, c):
        """Release the lock on the currently-locked device."""
        dev = self.selectedDevice(c)
        dev.unlock(c.ID)




import Queue

import pytest

import labrad
from labrad.server import LabradServer, setting, Signal
import labrad.util


class SignalTestServer(LabradServer):
    name = "Signal Test Server"

    class_signal = Signal(100, 'class_signal', '?')

    def __init__(self):
        super(SignalTestServer, self).__init__()
        self.instance_signal = Signal(200, 'instance_signal', '?')

    @setting(101)
    def fire_class_signal(self, c, data):
        self.class_signal(data)

    @setting(201)
    def fire_instance_signal(self, c, data):
        self.instance_signal(data)


def _test_signals(signal_setting, fire_setting):
    with labrad.util.syncRunServer(SignalTestServer()):
        with labrad.connect() as cxn:
            server = cxn.signal_test_server
            msg_id = 12345

            # add a listener to enqueue all messages
            queue = Queue.Queue()
            def handler(message_ctx, msg):
                queue.put((message_ctx, msg))
            cxn._backend.cxn.addListener(handler, ID=msg_id)

            # we don't get messages before signing up
            server[fire_setting]('not listening')
            with pytest.raises(Queue.Empty):
                queue.get(block=True, timeout=1)

            server[signal_setting](msg_id)
            server[fire_setting]('listening')
            msg_ctx, msg = queue.get(block=True, timeout=1)
            assert msg_ctx.source == server.ID
            assert msg == 'listening'

            server[signal_setting]()
            server[fire_setting]('not listening')
            with pytest.raises(Queue.Empty):
                queue.get(block=True, timeout=1)


def test_signal_can_be_defined_on_server_class():
    _test_signals('class_signal', 'fire_class_signal')


def test_signal_can_be_defined_on_server_instance():
    _test_signals('instance_signal', 'fire_instance_signal')


if __name__ == '__main__':
    pytest.main(['-v', __file__])



# Copyright (C) 2007  Matthew Neeley
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import operator

def gcd(a, b):
    """Compute the greatest common divisor of two ints."""
    a, b = b, a % b
    if b == 0:
        return a
    else:
        return gcd(a, b)

class Ratio(object):
    """Object representing the ratio of two integers."""

    def __init__(self, num, denom=1):
        if isinstance(num, Ratio):
            self.num = num.num
            self.denom = num.denom
        else:
            self.num = num
            self.denom = denom
            self._reduce()

    def _reduce(self):
        d = gcd(self.num, self.denom)
        self.num /= d
        self.denom /= d
        # ensure denominator is positive
        if self.denom < 0:
            self.num *= -1
            self.denom *= -1

    def _cmp(self, other, op):
        if not isinstance(other, (int, long, float, Ratio)):
            return NotImplemented
        if not isinstance(other, Ratio):
            other = Ratio(other)
        return op(self.num * other.denom, other.num * self.denom)

    def __lt__(self, other): return self._cmp(other, operator.lt)
    def __le__(self, other): return self._cmp(other, operator.le)
    def __eq__(self, other): return self._cmp(other, operator.eq)
    def __ne__(self, other): return self._cmp(other, operator.ne)
    def __gt__(self, other): return self._cmp(other, operator.gt)
    def __ge__(self, other): return self._cmp(other, operator.ge)

    def isInteger(self):
        return abs(self.denom) == 1

    def __int__(self):
        if not self.isInteger():
            raise TypeError("Cannot convert %r to int." % self)
        return int(self.num)

    def __long__(self):
        if not self.isInteger():
            raise TypeError("Cannot convert %r to long." % self)
        return long(self.num)

    def __float__(self):
        return float(self.num) / float(self.denom)

    def __nonzero__(self):
        return self.num != 0

    def __repr__(self):
        return 'Ratio(%s,%s)' % (self.num, self.denom)

    def __str__(self):
        if self.denom == 1:
            return '%s' % self.num
        else:
            return '%s/%s' % (self.num, self.denom)

    def __neg__(self):
        return Ratio(-self.num, self.denom)

    def __pos__(self):
        return self

    def __abs__(self):
        return Ratio(abs(self.num), abs(self.denom))

    def __add__(self, other):
        if not isinstance(other, (int, long, Ratio)):
            return NotImplemented
        if not isinstance(other, Ratio):
            other = Ratio(other)
        num = self.num * other.denom + self.denom * other.num
        denom = self.denom * other.denom
        return Ratio(num, denom)

    __radd__ = __add__

    def __iadd__(self, other):
        sum = self + other
        self.num = sum.num
        self.denom = sum.denom

    def __sub__(self, other):
        return self + (-other)

    def __rsub__(self, other):
        return other + (-self)

    def __isub__(self, other):
        self.__iadd__(self, -other)

    def __mul__(self, other):
        if not isinstance(other, (int, long, Ratio)):
            return NotImplemented
        if not isinstance(other, Ratio):
            other = Ratio(other)
        num = self.num * other.num
        denom = self.denom * other.denom
        return Ratio(num, denom)

    __rmul__ = __mul__

    def __imul__(self, other):
        prod = self * other
        self.num = prod.num
        self.denom = prod.denom

    def __div__(self, other):
        if not isinstance(other, (int, long, Ratio)):
            return NotImplemented
        if not isinstance(other, Ratio):
            other = Ratio(other)
        num = self.num * other.denom
        denom = self.denom * other.num
        return Ratio(num, denom)

    __truediv__ = __div__

    def __rdiv__(self, other):
        if not isinstance(other, (int, long, Ratio)):
            return NotImplemented
        if not isinstance(other, Ratio):
            other = Ratio(other)
        num = other.num * self.denom
        denom = other.denom * self.num
        return Ratio(num, denom)

    def __idiv__(self, other):
        ratio = self / other
        self.num = ratio.num
        self.denom = ratio.denom



# Copyright (C) 2007  Matthew Neeley
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import copy, re, textwrap
import contextlib
import os
import sys

from twisted.internet import defer, reactor
from twisted.internet.defer import inlineCallbacks, returnValue
from twisted.internet.threads import blockingCallFromThread
from twisted.python import log, reflect, util

from labrad import constants as C, crypto, support, thread
from labrad.support import getNodeName
from labrad.util.unwrap import unwrap

def fancyHelp(ID, name, accepts, returns, units, about):
    """Create informative help text for a server setting.
    """
    accepts = createTypeStr(accepts)
    returns = createTypeStr(returns)

    helptext = """
%(ID)d - %(name)s

ACCEPTED TYPES:
%(accepts)s

RETURNED TYPES:
%(returns)s

UNITS:
%(units)s

NOTES:
%(about)s""" % {'ID': ID, 'name': name, 'accepts': accepts,
                'returns': returns, 'units': units, 'about': about}
    return helptext

def createTypeStr(typelist):
    """Create a text list of accepted and returned types."""
    if len(typelist) == 0:
        return 'Any'
    return '[%s]' % ', '.join(descStr(item) for item in typelist)

def descStr(item):
    if isinstance(item, str):
        return item
    elif isinstance(item, tuple):
        return '%s (%s)' % (item[0].__name__, item[1])
    else:
        return item.__name__

DOC_PATTERN = re.compile(r'(.*)NOTES?:\s*(.*)', re.DOTALL)
def parseSettingDoc(s):
    if s is None:
        return '', ''
    lines = s.split('\n')
    first, rest = lines[0].strip(), lines[1:]
    rest = textwrap.dedent('\n'.join(rest))
    s = first + '\n' + rest
    try:
        descr, notes = DOC_PATTERN.search(s).groups()
    except:
        descr, notes = s, ''
    return unwrap(descr.strip()), unwrap(notes)


# a list of printable representations of the ascii character codes
FILTER = ''.join(chr(x) if len(repr(chr(x))) == 3 else '.' for x in range(256))

def dump(src, length=16):
    """Nicely-formatted hex dump of raw data."""
    N = 0; result = ''
    while src:
        s, src = src[:length], src[length:]
        hexa = ' '.join('%02X' % ord(x) for x in s)
        s = s.translate(FILTER)
        result += '%04X   %-*s   %s\n' % (N, length*3, hexa, s)
        N += length
    return result

def linspace(star, stop, N):
    """Linearly-spaced list of numbers."""
    return [star + (stop - star) * n / (N-1) for n in range(N)]

class ContextDict(dict):
    """Subclass of dict for holding context data.

    Using a subclass allows us to set attributes: ID and source.
    """

def findEnvironmentVars(string):
    """Find all environment variables of the form %VAR% in a string."""
    return re.findall('%([^%]+)%', string)

def interpEnvironmentVars(string, env=None):
    """Replace all environment variables of the form %VAR% in a string.

    Values are taken from the env variable, a dict-like object.  Variable
    names are converted to upper case before interpolation, to maintain
    case insensitivity.  If any required variables are not found in env,
    this raises a KeyError.
    """
    if env is None:
        import os
        env = os.environ
    for label in findEnvironmentVars(string):
        tag = '%' + label + '%'
        string = string.replace(tag, env[label.upper()])
    return string

def timedeltaToSeconds(td):
    return td.seconds + td.microseconds / 1.0e6

def timing(f, n=100, **kw):
    from datetime import datetime

    total = 0
    for _ in xrange(n):
        start = datetime.now()
        f(**kw)
        end = datetime.now()
        delta = end - start
        total += timedeltaToSeconds(delta)
    avg = total / float(n)
    return avg



# convenience functions for dealing with units
def convert(v, u):
    """Convert quantity v into units u.

    If v is just a number, no conversion is performed.
    If u is None, then the units of v are simply stripped
    but otherwise no conversion is performed.
    """
    if hasattr(v, 'value'): # prefer over subclass check: isinstance(v, Value)
        if u is None:
            return v.value
        else:
            return v[u]
    else:
        return v

def convertUnits(**unitdict):
    """
    Decorator to create functions that automatically
    convert arguments into specified units.  If a unit
    is specified for an argument and the caller passes
    an argument with incompatible units, an Exception
    will be raised.  Inside the decorated function, the
    arguments no longer have any units, they are just
    plain floats.  Not all arguments to the function need
    to be specified in the decorator.  Those that are not
    specified will be passed through unmodified.

    Usage:

    @convertUnits(t0='ns', amp=None)
    def func(t0, amp):
        <do stuff>

    This is essentially equivalent to:

    def func(t0, amp):
        t0 = convert(t0, 'ns')
        amp = convert(amp, None)
        <do stuff>

    The convert function, defined above, will convert
    any quantities with units into the specified units,
    or strip off any units if unit is None.
    """

    def wrap(f):
        args = inspect.getargspec(f)[0] # list of argument names
        for arg in unitdict:
            if arg not in args:
                raise Exception('function %s does not take arg "%s"' % (f, arg))
        # unitdict maps argument names to units
        # posdict maps argument positions to units
        posdict = dict((i, unitdict[arg]) for i, arg in enumerate(args) if arg in unitdict)

        @functools.wraps(f)
        def wrapped(*a, **kw):
            # convert positional arguments if they have a unit
            a = [convert(val, posdict.get(i, None)) for i, val in enumerate(a)]
            # convert named arguments if they have a unit
            for arg, val in kw.iteritems():
                if arg in unitdict:
                    kw[arg] = convert(val, unitdict[arg])
            # call the function with converted arguments
            return f(*a, **kw)
        return wrapped
    return wrap


def is_local_connection(transport):
    """Determine whether the given network transport is connected to localhost.

    Args:
        transport (twisted.internet.interfaces.ITransport): a twisted transport
            object, representing a network connection.

    Returns (boolean):
        True if the transport is connected to a peer on the local host, False
        otherwise.
    """
    remote_addr = transport.getPeer().host
    local_addr = transport.getHost().host
    return remote_addr in ['127.0.0.1', '::1'] or remote_addr == local_addr


# deferred helpers

def wakeupCall(delay, args=None):
    """Return a deferred that will be called back
    after a specified delay.
    """
    d = defer.Deferred()
    reactor.callLater(delay, d.callback, args)
    return d

def firstToFire(n=2):
    heads = [defer.Deferred() for _ in range(n)]
    first = defer.DeferredList(heads, fireOnOneCallback=True,
                                      fireOnOneErrback=True,
                                      consumeErrors=True)
    first.addCallback(lambda result: result[0])
    return first, heads

@inlineCallbacks
def maybeTimeout(deferred, timeout, timeoutResult):
    """Takes a deferred and returns a new deferred that might timeout."""
    td = wakeupCall(timeout, timeoutResult)
    result, _index = yield defer.DeferredList([deferred, td],
                                              fireOnOneCallback=True,
                                              fireOnOneErrback=True,
                                              consumeErrors=True)
    returnValue(result)

class DeferredSignal(object):
    """An object that can create multiple deferreds on demand.

    When the signal is fired, all created deferreds will be fired
    or have their errback method called, as appropriate.
    """
    def __init__(self):
        self.waiters = []
        self.listeners = []
        self.fired = None

    def callback(self, data=None):
        self._fire(True, data)

    def errback(self, reason=None):
        self._fire(False, reason)

    def _fire(self, success, data):
        self.fired = success, data
        waiters = self.waiters
        self.waiters = []
        for d in waiters:
            if success:
                d.callback(data)
            else:
                d.errback(data)
        if success:
            for func in self.listeners:
                func(data)

    def __call__(self):
        if self.fired:
            success, data = self.fired
            if success:
                return defer.succeed(data)
            else:
                return defer.fail(data)
        else:
            d = defer.Deferred()
            self.waiters.append(d)
            return d

    def connect(self, listener):
        if listener not in self.listeners:
            self.listeners.append(listener)

    def disconnect(self, listener):
        self.listeners.remove(listener)


# run labrad server

def parseServerOptions(name, exit_on_failure=True, options=None):
    """Parse standard command line options for a server.

    Args:
        name (string): The default server name to use if no name is specified
            on the command line.
        exit_on_failure (boolean): If True, we call sys.exit when we fail
            to parse the command line options. Otherwise, we raise UsageError.
        options (list(string)): If given, parse options from the given strings.
            Otherwise, will parse options from the command line in sys.argv.

    Returns:
        A ServerOptions instance initialized from the command line arguments.
        This is a dict-like object containing these string-valued keys:
            'name', 'node', 'host', 'port', 'password', 'tls'
    """
    from twisted.python import usage

    class ServerOptions(usage.Options):
        optParameters = [
            ['name', 'n', name, 'Server name.'],
            ['node', 'd', getNodeName(), 'Node name.'],
            ['host', 'h', C.MANAGER_HOST, 'Manager location.'],
            ['port', 'p', None, 'Manager port.', int],
            ['username', 'u', None, 'Username.'],
            ['password', 'w', None, 'Login password.'],
            ['tls', 's', C.MANAGER_TLS,
             'TLS mode for connecting to manager (on/starttls/off)']]

    config = ServerOptions()
    config['tls'] = C.check_tls_mode(config['tls'])
    try:
        config.parseOptions(options=options)
        if config['port'] is None:
            tls_on = config['tls'] == 'on'
            config['port'] = C.MANAGER_PORT_TLS if tls_on else C.MANAGER_PORT
    except usage.UsageError, errortext:
        print '%s: %s' % (sys.argv[0], errortext)
        print '%s: Try --help for usage details.' % (sys.argv[0])
        if exit_on_failure:
            sys.exit(1)
        else:
            raise
    return config

def updateServerOptions(srv, config):
    """Update server options (name, instanceName, password) from cmdline config.

    Args:
        srv: An instance of LabradServer to be updated to match the config.
        config: A dict-like object containing keys 'name', 'instanceName' and
            'password'. Typically, this will be a ServerOptions instance
            returned by parseServerOptions.
    """
    if config['name']:
        srv.name = config['name']
    env = dict(os.environ)
    env['LABRADNODE'] = config['node']
    srv.name = interpEnvironmentVars(srv.name, env)
    if hasattr(srv, 'instanceName'):
        srv.instanceName = interpEnvironmentVars(srv.instanceName, env)

def runServer(srv, run_reactor=True, stop_reactor=True):
    """Run the given server instance.

    Args:
        srv (LabradServer): The server instance to run.
        run_reactor (boolean): If True, we call reactor.run() to start the
            twisted reactor. Otherwise, the caller must ensure that reactor.run
            is called.
        stop_reactor (boolean): If True, we call reactor.stop() after the
            server instance shuts down (whether normally or due to an error
            condition. Otherwise, the caller must arrange to call reactor.stop.
    """
    from labrad import protocol

    config = parseServerOptions(name=srv.name)
    updateServerOptions(srv, config)

    log.startLogging(sys.stdout)
    #observer = MyLogObserver(sys.stdout)
    #log.startLoggingWithObserver(observer.emit)

    @inlineCallbacks
    def run(srv):
        host = config['host']
        port = int(config['port'])
        tls_mode = config['tls']
        try:
            p = yield protocol.connect(host, port, tls_mode)
            yield p.authenticate(config['username'], config['password'])
            yield srv.startup(p)
            yield srv.onShutdown()
            log.msg('Disconnected cleanly.')
        except Exception as e:
            log.msg('There was an error: {}'.format(e))
        if stop_reactor:
            try:
                reactor.stop()
            except Exception:
                pass

    _ = run(srv)
    if run_reactor:
        reactor.run()

@contextlib.contextmanager
def syncRunServer(srv, host=C.MANAGER_HOST, port=None, username=None,
                  password=None, tls_mode=C.MANAGER_TLS):
    """Run a labrad server of the specified class in a synchronous context.

    Returns a context manager to be used with python's with statement that
    will yield when the server has started and then shut the server down after
    the context is exited.
    """
    from labrad import protocol

    tls_mode = C.check_tls_mode(tls_mode)

    if port is None:
        port = C.MANAGER_PORT_TLS if tls_mode == 'on' else C.MANAGER_PORT

    @inlineCallbacks
    def start_server():
        p = yield protocol.connect(host, port, tls_mode)
        yield p.authenticate(username, password)
        yield srv.startup(p)

    @inlineCallbacks
    def stop_server():
        srv.disconnect()
        yield srv.onShutdown()

    thread.startReactor()
    blockingCallFromThread(reactor, start_server)
    try:
        yield
    finally:
        try:
            blockingCallFromThread(reactor, stop_server)
        except Exception:
            pass # don't care about exceptions here



class MyLogObserver(log.FileLogObserver):
    timeFormat = '%Y/%m/%d %H:%M -'

    def emit(self, eventDict):
        edm = eventDict['message']
        if not edm:
            if eventDict['isError'] and 'failure' in eventDict:
                text = ((eventDict.get('why') or 'Unhandled Error')
                        + '\n' + eventDict['failure'].getTraceback())
            elif 'format' in eventDict:
                text = self._safeFormat(eventDict['format'], eventDict)
            else:
                # we don't know how to log this
                return
        else:
            text = ' '.join(map(reflect.safe_str, edm))

        timeStr = self.formatTime(eventDict['time'])
        fmtDict = {'text': text.replace("\n", "\n\t")}
        msgStr = self._safeFormat("%(text)s\n", fmtDict)

        util.untilConcludes(self.write, timeStr + " " + msgStr)
        util.untilConcludes(self.flush)  # Hoorj!




# Copyright (C) 2007  Matthew Neeley
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
labrad module for python.
"""

from labrad import backend, client, constants

def connect(host=constants.MANAGER_HOST, port=None, name=None, **kw):
    """Create a client connection to the labrad manager."""
    cxn = backend.connect(host=host, port=port, name=name, **kw)
    return client.Client(cxn)

connection = connect



# Copyright (C) 2007  Matthew Neeley
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from labrad.server import LabradServer

class ServerTemplate(LabradServer):
    """Template for a python LabRAD server.

    This doc string will appear as descriptive help text when
    this server connects to LabRAD.
    """

    # the server name will identify this server in LabRAD
    # note that this name should match that given in the
    # .ini file for this server, if you want to use this
    # server with the node.
    name = 'Python Server Template'

    # Server startup and shutdown
    #
    # these functions are called after we first connect to
    # LabRAD and before disconnecting.  Here you should perform
    # any global initialization or cleanup.
    def initServer(self):
        """Initialize Server.

        Called after registering settings and creating a client
        connection to labrad, but before we start serving requests.
        """

    def stopServer(self):
        """Stop the server.

        Called when the server is shutting down, but before we have
        closed any client connections.  Perform any cleanup operations here.
        """

    # Context handling
    #
    # these functions are called to initialize and expire
    # context objects for requests.  The object c is a
    # dict-like object that you can use to store and data you
    # would like to have associated with the context.  This same
    # context object will get passed in to request handlers as well.
    def initContext(self, c):
        """Initialize a new context object."""

    def expireContext(self, c):
        """Expire Context.

        Called when a client which created a context disconnects,
        or when the client explicitly requests the expiration.
        Any cleanup operations on the context should be done here.
        """

    # Manager notifications
    #
    # These functions will be called in response to notifications
    # from the manager about events happening in the LabRAD system.
    # They let you respond to the connection and disconnection
    # of other servers.
    def serverConnected(self, ID, name):
        """This function will be called when a new server connects to LabRAD."""

    def serverDisconnected(self, ID, name):
        """This function will be called when a server disconnects from LabRAD."""


# create an instance of our server class
__server__ = ServerTemplate()

# this is some boilerplate code to run the
# server when this module is executed
if __name__ == '__main__':
    from labrad import util
    util.runServer(__server__)



"""
### BEGIN NODE INFO
[info]
name = Python Threaded Test Server
version = 1.0.0
description = Basic python server.

[startup]
cmdline = %PYTHON% %FILE%
timeout = 20

[shutdown]
message = 987654321
timeout = 5
### END NODE INFO
"""

import time

from labrad import types as T, util
from labrad.server import ThreadedServer, setting
from labrad.units import m, s
from labrad.util import hydrant


class TestServer(ThreadedServer):
    """Server to test labrad from python.

    This server provides a number of settings that
    are useful for testing labrad code, network profiling,
    and testing the python labrad library.
    """
    name = 'Python Threaded Test Server'
    testMode = True
    shutdownMessage = 987654321

    def initServer(self):
        time.sleep(0.5)

    def stopServer(self):
        print 'before sleep'
        time.sleep(0.5)
        print 'after sleep'

    def serverConnected(self, ID, name):
        print 'server connected:', ID, name

    def serverDisconnected(self, ID, name):
        print 'server disconnected:', ID, name

    def initContext(self, c):
        c['delay'] = 1*s
        c['dict'] = {}

    @setting(2, "Delayed Echo", data='?')
    def delayed_echo(self, c, data):
        """Echo a packet after a specified delay."""
        time.sleep(c['delay'][s])
        return data

    @setting(3, "Future Echo", data='?')
    def future_echo(self, c, data):
        """Echo a packet after a specified delay."""
        f = self.client.manager.echo.future(data)
        time.sleep(c['delay'][s])
        return f

    @setting(4, "Echo Delay", delay='v[s]', returns='v[s]')
    def echo_delay(self, c, delay=None):
        """Get or set the echo delay."""
        self.log('Echo delay: %s' % delay)
        if delay is not None:
            c['delay'] = delay
        return c['delay']

    @setting(2000, "Registry Dir", path='*s', returns='(*s, *s)')
    def registry_dir(self, c, path=['']):
        """Get Registry listing for the specified path."""
        p = self.client.registry.packet()
        p.cd([''])
        p.cd(path)
        p.dir()
        result = p.send()
        dirs, keys = result['dir']
        return dirs, keys

    @setting(42, "Delayed Echo Wrapper", data='?', returns='?')
    def delayed_echo_wrapper(self, c, data):
        """Echo a packet after a delay just like delayed_echo.

        This tests calling a setting from another setting.
        """
        rv = self.delayed_echo(c, data)
        return rv

    @setting(40, "Speed", speed='v[m/s]', returns='v[m/s]')
    def speed(self, c, speed=None):
        """Get or set the speed."""
        self.log('Speed: %s' % speed)
        if speed is not None:
            c['speed'] = speed
        return c['speed']

    @setting(41, "Verbose Echo", data='?')
    def verbose_echo(self, c, data):
        print type(data)
        print repr(data)
        return data

    @setting(5, "Exc in Handler", data='?')
    def exc_in_handler(self, c, data):
        """Raises an exception directly in the handler."""
        self.log('Exception in handler.')
        raise Exception('Raised in handler.')

    @setting(9, "Exc after delay", data='?')
    def exc_in_inlinecallback(self, c, data):
        """Raises an exception after a delay."""
        self.log('Exception after delay.')
        time.sleep(c['delay'][s])
        raise Exception('Raised after delay.')

    @setting(10, "Bad Return Type", data='?', returns='s')
    def bad_return_type(self, c, data):
        """Returns a value that does not match the declared return type."""
        return 5

    @setting(11, "Get Random Data", tag='s')
    def get_random_data(self, c, tag=None):
        """Get a random bit of data conforming to the specified type tag."""
        if tag is None:
            t = hydrant.randType()
        else:
            t = T.parseTypeTag(tag)
        return hydrant.randValue(t)

    @setting(12, "Get Random Tag")
    def get_random_tag(self, c):
        """Get a random LabRAD type tag."""
        return str(hydrant.randType())

    @setting(100, "Set", unflatten=False, key='s', value='?', returns='')
    def set(self, c, key, value):
        c['dict'][key.unflatten()] = value

    @setting(101, "Get", key='s', returns='?')
    def get(self, c, key):
        print "getting tag: %s, value: %s" % (key, c['dict'][key])
        return c['dict'][key]

    @setting(102, "Keys", returns='*s')
    def keys(self, c):
        return sorted(c['dict'].keys())

    @setting(103, "Set Reversed", value='?', key='s', unflatten=False)
    def set_reversed(self, c, value, key):
        """Same as self.set() with argument order reversed."""
        self.set(c, key, value)

    @setting(1000, path=['', 's'], file=['s'])
    def first_arg_can_have_default_value(self, c, path=None, file=None):
        """Exercise setting decorator fix from issue #242."""
        pass


if __name__ == '__main__':
    util.runServer(TestServer())



import threading

import pytest

from labrad.server import LabradServer
from labrad import util

def test_sync_run_server():
    """Ensure that syncRunServer runs the stopServer method on shutdown."""
    event = threading.Event()

    class TestServer(LabradServer):
        name = "TestServer"

        def stopServer(self):
            event.set()

    server = TestServer()
    with util.syncRunServer(server):
        pass
    if not event.wait(0.5):
        raise Exception('event not set in stopServer method')

if __name__ == '__main__':
    pytest.main(['-v', __file__])



''' unwrap.py -- a function that makes every paragraph in a text into a single long line
                  so that the text will look right in a "word wrapping" environment
                  (e.g. a word processor, HTML text box or a Palm Doc document).

    A paragraph is:
        1. an indented line followed by one or more unidented lines;
        2. two or more unidented lines with a blank line above.

    An indented line contains text and has leading spaces.
    An unindented line contains text but has no leading spaces.
    A blank line may contain spaces but not text; it may also be the start of the text.

    Anything other than those two types of paragraph (e.g. blocks of
    fully indented text, single lines) will be left alone.

        @    .  .
    glyn wave co nz
'''
#
# LICENCE
#     This file is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2, or (at your option)
#     any later version.
#
#     This file is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have a several hundred copies of of the GNU General Public
#     License on your hard drive.  If not, install Linux, read
#     http://www.gnu.org/copyleft/gpl.html or write to the Free Software
#     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

__all__ = ['unwrap']

import re

def unwrap(text):
    ''' unwrap (text : string) -> string
        A function that makes every paragraph in a text into a single long line
        so that the text will look right in a "word wrapping" environment
        (e.g. a word processor, HTML text box or a Palm Doc document). '''
    return paragraph.sub(unwrap_para, text)

def regex(s):
    ''' Compile a regex using a "little language" of symbolic names. '''
    for name, expr in names.items():
        s = s.replace(name, expr)
    return re.compile(s, re.DOTALL | re.MULTILINE | re.VERBOSE)

names = {'space':    r'[ \t]',
         'text':     r'[^ \t\n] [^\n]*',
         'newline':  r'\n'
     }

paragraph = regex('''
  ^ ( space+ ) ( text (?: newline text )+ ) $               # indented
| ^ ( space* newline | ^) ( text (?: newline text )+ ) $    # blank-line seperated
''')

whitespace = re.compile(r'\s+', re.DOTALL)

def unwrap_para(match):
    if match.group(3) is None:
        return match.group(1) + whitespace.sub(' ', match.group(2))
    else:
        return match.group(3) + whitespace.sub(' ', match.group(4))

#end



#!/usr/bin/python
"""Measure some performance figures on the unit library

run_tests will run every function named perf_X, and create
a dictionary of the timing results.  perf_ tests can have
an optional argument 'N' which is a number of iterations
to run.
"""
import argparse
import json
import random
import timeit
import xml.etree.ElementTree as ET

import numpy as np

import labrad.units as U


unit_list = [x for x in U._unit_table.values() if x.offset == 0]


def random_unit():
    return random.choice(unit_list)


def perf_mul_float(N=1000):
    a = 1.0
    b = 230.0
    for j in range(N):
        c = a * b


def perf_mul_value_random(N=1000):
    for j in range(N):
        u1 = random_unit()
        u2 = random_unit()
        c = (1.0*u1) * (1.0*u2)


def perf_mul_unit_random(N=1000):
    for j in range(N):
        u1 = random_unit()
        u2 = random_unit()
        up = u1 * u2


def perf_mul_unit_cached(N=1000):
    u1 = random_unit()
    u2 = random_unit()
    for j in range(N):
        up = u1 * u2


def perf_mul_value_cached(N=1000):
    u1 = random_unit()
    u2 = random_unit()
    for j in range(N):
        c = (1.0*u1) * (1.0*u2)


def perf_add_value_equal(N=1000):
    u1 = U.m
    for j in range(N):
        c = (1*u1) + (1*u1)


def perf_add_value_unequal(N=1000):
    u1 = U.m
    u2 = U.nm
    for j in range(N):
        c = (1*u1) + (1*u2)


def perf_envelope_unit(N=1000):
    t = np.arange(20*N)*U.ns
    t_cos = np.arange(20)*U.ns
    w_cos = np.pi * 2 * U.GHz / 20
    z = np.zeros(20*N, dtype=np.complex128)
    for j in range(N):
        w = random.random() * .1 * U.GHz
        phi = random.random()*np.pi
        z[j*20:(j+1)*20] = (1-np.cos(t_cos*w_cos))*np.exp(1j*w*t[j*20:(j+1)*20]+phi)
    return z


def perf_envelope_no_unit(N=1000):
    t = np.arange(20*N)
    t_cos = np.arange(20)
    w_cos = np.pi * 2 / 20
    z = np.zeros(20*N, dtype=np.complex128)
    for j in range(N):
        w = random.random() * .1
        phi = random.random() * np.pi
        z[j*20:(j+1)*20] = (1-np.cos(t_cos*w_cos))*np.exp(1j*w*t[j*20:(j+1)*20]+phi)
    return z


def perf_mul_array_unit(N=1000):
    a1 = np.arange(1000)*U.ns
    a2 = np.arange(1000)*U.GHz
    for j in range(N):
        a3 = a1 * a2


def perf_mul_array(N=1000):
    a1 = np.arange(1000)
    a2 = np.arange(1000)
    for j in range(N):
        a3 = a1 * a2


def run_tests(func_prefix='perf_', key_prefix='labrad.units.'):
    stats = {}
    g = globals()
    for name in g.keys():
        if name.startswith(func_prefix):
            key = key_prefix + name[len(func_prefix):]
            result = timeit.repeat(g[name], number=10, repeat=3)
            stats[key] = min(result)/10000 * U.s
    return stats


def main():
    parser = argparse.ArgumentParser(description='Run unit tests')
    parser.add_argument('--format', nargs='?',
                        choices=['xml', 'json'], default='json',
                        help='ouptut format in which to report results')

    args = parser.parse_args()

    result = run_tests()

    if args.format == 'json':
        print json.dumps({k: v['ms'] for k, v in result.items()}, indent=2)
    else:
        root = ET.Element('build')
        for k, v in sorted(result.items()):
            child = ET.SubElement(root, 'statisticValue',
                                  key=k, value='{:f}'.format(v['ms']))
        print ET.tostring(root)


if __name__ == "__main__":
    main()



import base64
import hashlib
import logging
import os
import re


try:
    import twisted
    if int(twisted.__version__.split('.')[0]) >= 14:
        from twisted.internet import ssl
        TLS = True
    else:
        logging.warning("Twisted version >= 14.0.0 required for SSL support. "
                        "Older versions don't support the platform CA roots.")
        TLS = False
except ImportError:
    logging.warning('pyOpenSSL not found. Without encryption you will only be '
                    'able to connect to the labrad manager on localhost.')
    TLS = False


# Location where we store trusted self-signed manager certs.
# The manager cert file for a given hostname is named <hostname>.cert
# Note that real certs signed by public CAs are not stored here,
# because we can validate them with the system's trusted certificates.
CERTS_PATH = os.path.expanduser('~/.labrad/client/certs')


def tls_options(hostname, cert_string=None):
    """Get certificate options for TLS connections.

    If we are unable to use TLS because pyOpenSSL is not installed, this will
    raise an exception.

    Args:
        hostname (string): The hostname we are trying to connect to with TLS.
        cert_string (string): A certificate in PEM format. If provided, we
            return CertificateOptions that will trust only the given cert
            (assuming its canonical name attribute matches the given hostname).
            If not provided, we will try to load a suitable certificate for
            this hostname from our store of trusted self-signed certificates,
            and otherwise fall back to just using the trust roots configured
            for the OS to validate the server certificate.

    Returns (twisted.internet.ssl.CertificateOptions):
        Certificate suitable for use by TLS clients to connect to the given

    Raises:
        Exception: could not import pyOpenSSL package needed for TLS.
    """
    if not TLS:
        raise Exception('Unable to connect to labrad with TLS. Please be sure '
                        'to install pyOpenSSL.')
    if cert_string is None:
        cert_string = load_cert(hostname)
    if cert_string is None:
        logging.info('No certificate found for host "{}" in {}. Will use '
                     'system certs to verify tls'.format(hostname, CERTS_PATH))
        return ssl.optionsForClientTLS(hostname.decode('utf-8'))
    else:
        cert = ssl.Certificate.loadPEM(cert_string)
        return ssl.optionsForClientTLS(hostname.decode('utf-8'), cert)


def save_cert(hostname, cert_string):
    """Save the given self-signed certificate as a trusted certificate.

    Args:
        hostname (string): The hostname for this certificate. Must match the
            canonical name attribute (CN) in the certificate itself. We do not
            check this here, but if it is not the case, TLS negotiation using
            this cert will fail.
        cert_string (string): Certificate in PEM format.
    """
    # sanity check that this is a valid PEM string
    cert = ssl.Certificate.loadPEM(cert_string)

    fname = hostname + '.cert'
    full_name = os.path.join(CERTS_PATH, fname)
    if not os.path.exists(CERTS_PATH):
        os.makedirs(CERTS_PATH)
    with open(full_name, 'w') as f:
        f.write(cert_string)


def load_cert(hostname):
    """Load a trusted certificate for the given hostname.

    Args:
        hostname(string): The hostname for which we want a certificate.

    Returns (string | None):
        A certificate for the specified hostname as a PEM-encoded string. If no
        certificate is found for the given hostname, we return None.
    """
    if not os.path.exists(CERTS_PATH):
        return None
    for fname in os.listdir(CERTS_PATH):
        full_name = os.path.join(CERTS_PATH, fname)
        if not os.path.isfile(full_name):
            continue
        host, ext = os.path.splitext(fname)
        if host == hostname and ext == '.cert':
            with open(full_name) as f:
                return f.read()
    else:
        return None


def fingerprint(cert_string, digest=hashlib.sha1):
    """Get a human-readable digest of the given certificate string.

    Certificate fingerprints are computed from the DER-encoded (binary) form of
    the certificate, while PEM is a text format containing the base64-encoded
    DER bytes. So we strip off boilerplate/whitespace and base64-decode before
    computing the digest.

    Args:
        cert_string (string): PEM-encoded certificate.
        digest: the hashlib algorithm to use; a callable that returns an
            instance of a message digest algorithm.

    Returns (string):
        Human-readable hex version of the certificate digest in the form:
        'AA:BB:CC:...:FF'.
    """
    BEGIN_CERT = '-----BEGIN CERTIFICATE-----'
    END_CERT = '-----END CERTIFICATE-----'
    begin_idx = cert_string.index(BEGIN_CERT) + len(BEGIN_CERT)
    end_idx = cert_string.index(END_CERT)
    cert_b64 = re.sub('\s+', '', cert_string[begin_idx:end_idx])
    cert_bytes = base64.b64decode(cert_b64)
    m = digest()
    m.update(cert_bytes)
    digest = m.hexdigest()

    def grouped(s, n=2):
        while s:
            yield s[:n]
            s = s[n:]

    return ':'.join(g.upper() for g in grouped(digest, 2))



# test to benchmark the parsing of lists of string data
# this revealed a serious inefficiency in the implementation
# of labrad.types.Buffer, which is used in the parsing process.
# fixing the implementation resulted in a substantial performance gain.

import time

import numpy as np

from labrad import types

def timeIt(f, *a, **kw):
    start = time.time()
    result = f(*a, **kw)
    end = time.time()
    print f.__name__, '- elapsed:', end - start
    return result

def extractAverage(packets):
    data = ''.join(packets)
    Is, Qs = np.fromstring(data, dtype='<i2').reshape(-1, 2).astype(int).T
    return (Is, Qs)

def extract(packets):
    result = [data for src, dest, eth, data in packets]
    answer = extractAverage(result)
    return answer

mac = '01:23:45:67:89:ab'
eth = 1

packets = [(mac, mac, eth, '\x00'*44) for _ in range(9000)]
data, t, endianness = types.flatten(packets)

timeIt(extract, packets)
data, t, endianness = timeIt(types.flatten, packets)
timeIt(types.unflatten, data, t, endianness)



# Copyright (C) 2007  Matthew Neeley
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

import functools
from types import MethodType

from twisted.internet import defer
from twisted.internet.defer import inlineCallbacks, returnValue

from labrad import constants as C, manager, protocol, types as T
from labrad.support import (indent, mangle, extractKey, MultiDict, PacketRecord,
                            PacketResponse, hexdump)


class AsyncSettingWrapper(object):
    """Represents a setting on a remote LabRAD server."""

    def __init__(self, server, name, pyName, ID, info):
        self.name = self.__name__ = self._labrad_name = name
        self._py_name = pyName
        self._updateInfo(ID, info)
        self._server = server
        self._cxn = server._cxn
        self._prot = server._cxn._cxn
        self._num_listeners = 0

    def _updateInfo(self, ID, info):
        """Update meta information about this setting."""
        self.ID = ID
        self.__doc__, self.accepts, self.returns, self.notes = info

    @inlineCallbacks
    def __call__(self, *args, **kw):
        """Send a request to this setting."""
        tag = extractKey(kw, 'tag', None) or self.accepts
        if len(args) == 0:
            args = None
        elif len(args) == 1:
            args = args[0]
        flat = T.flatten(args, tag)
        r = yield self._server._send([(self.ID, flat)], **kw)
        returnValue(r[0][1])

    @inlineCallbacks
    def connect(self, handler, context=(0, 0),
                connectargs=(), connectkw={},
                handlerargs=(), handlerkw={}):
        """Connect a handler to messages from this signal.

        This is only applicable if the remote setting handles
        signal connection and disconnection.  We also keep
        track of the number of handlers that have been added to
        this setting.
        """
        self._prot.addListener(handler,
                               source=self._server.ID, context=context, ID=self.ID,
                               args=handlerargs, kw=handlerkw)
        self._num_listeners += 1
        try:
            yield self.__call__(self.ID, context=context, *connectargs, **connectkw)
        except Exception:
            self._prot.removeListener(handler,
                                      source=self._server.ID, context=context, ID=self.ID)
            self._num_listeners -= 1
            raise

    def disconnect(self, handler, context=(0, 0),
                   disconnectargs=(), disconnectkw={}):
        """Disconnect a handler for messages from this signal.

        If the number of handlers drops to zero, we make a request to
        tell the signal to stop sending us messages.
        """
        self._prot.removeListener(handler, source=self._server.ID, context=context, ID=self.ID)
        self._num_listeners = max(self._num_listeners - 1, 0)
        return self.__call__(context=context, *disconnectargs, **disconnectkw)

class SettingBinder(object):
    """A dictionary proxy that binds methods to an instance.

    This is used to emulate the old 'settings' attribute on a packet
    wrapper, now that the settings have been moved into a class.
    """
    # TODO fill out dict functionality here (.keys(), .items(), etc.)
    def __init__(self, inst):
        self._inst = inst

    def __getitem__(self, key):
        inst = self._inst
        return inst._bind(inst.__class__.settings[key])

class AsyncPacketWrapper(object):
    """Represents a LabRAD packet to a server.

    One wrapper class is created for each server.  Settings can be added or
    removed from the wrapper class as needed, if the server settings change.
    """

    @classmethod
    def _addSetting(cls, setting):
        """Add a new instance method for the specified setting.

        Instance methods are cached, so that if a setting goes away and
        comes back later, the method will still point back to the
        same setting object.
        """
        if setting.name in cls._cache:
            method = cls._cache[setting.name]
        else:
            def wrapped(self, *args, **kw):
                key = extractKey(kw, 'key', None)
                tag = extractKey(kw, 'tag', None) or setting.accepts
                if len(args) == 0:
                    args = None
                elif len(args) == 1:
                    args = args[0]
                flat = T.flatten(args, tag)
                rec = PacketRecord(ID=setting.ID, data=args, tag=tag, flat=flat,
                                   key=key, name=setting.name)
                self._packet.append(rec)
                return self
            wrapped.name = setting.name
            method = MethodType(wrapped, None, cls)
        cls._cache[setting.name] = method
        cls.settings[setting.name, setting._py_name, setting.ID] = method
        setattr(cls, setting._py_name, method)

    @classmethod
    def _refreshSetting(cls, setting):
        """Refresh a setting by updating its aliases in our setting MultiDict."""
        cls.settings._updateAliases(setting.name, setting._py_name, setting.ID)

    @classmethod
    def _delSetting(cls, setting):
        """Delete the instance method for the specified setting."""
        del cls.settings[setting.name]
        delattr(cls, setting._py_name)


    def __init__(self, **kw):
        """Create a new packet."""
        self._packet = []
        self._kw = kw
        self.settings = SettingBinder(self)

    @inlineCallbacks
    def send(self, **kw):
        """Send this packet to the server."""
        # drop keys from records before sending
        records = [(rec.ID, rec.flat) for rec in self._packet]
        r = yield self._server._send(records, **dict(self._kw, **kw))
        returnValue(PacketResponse(r, self._server, self._packet))

    def _bind(self, method):
        """Bind a method to this instance."""
        return functools.partial(method, self)

    def __getitem__(self, key):
        """Get a setting method from the class and bind it to this instance."""
        return self._bind(self.__class__.settings[key])

    def __setitem__(self, key, value):
        """Update existing parts of the packet, indexed by key.

        Note that if multiple records share the same key, they will
        all be updated.
        """
        for i, rec in enumerate(self._packet):
            if key == rec.key:
                flat = T.flatten(value, rec.tag)
                self._packet[i] = rec._replace(data=value, flat=flat)

    def __delitem__(self, key):
        """Delete a setting call from this packet, indexed by key.

        Note that if multiple records share the same key, they will
        all be deleted.
        """
        for i, rec in enumerate(self._packet):
            if key == rec.key:
                self._packet.pop(i)

    def _dataStr(self, data):
        if len(data) < 160:
            return self._dataRepr(data)
        else:
            x = self._dataRepr(data[0:32])
            '\n'.join('    ' + y for y in x.splitlines())
            return "string of length %d beginning with... \n%s" % (len(data), x)

    def _dataRepr(self, data):
        if all((ord(x) > 31 and ord(x) < 127) or x.isspace() for x in data): # Is string printable ascii
            return data
        else:
            return hexdump(data)

    def _recordRepr(self, rec, short=False):
        """Create a string representation of a packet record.

        Includes the setting name and result key (if specified), as well as
        a possibly truncated repr of the data for this setting.
        """
        key_str = "" if rec.key is None else " (key=%s)" % (rec.key,)
        prefix = '%s%s: ' % (self._server.settings[rec.ID].name, key_str)
        data_str = self._dataStr(str(rec.data)) if short else self._dataRepr(str(rec.data))
        data_str = data_str.replace('\n', '\n' + ' '*len(prefix))
        return prefix + data_str

    def __str__(self):
        return self.__repr__(True)

    def __repr__(self, short=False):
        """Create a string representation of this packet.

        Includes the server the packet is intended for, as well as all the
        records that have been added to the packet.
        """
        data_str = '\n'.join(self._recordRepr(rec, short=short) for rec in self._packet)
        return unwrap("""\
            |Packet for server: '%s'
            |
            |Data:
            |%s""") % (self._server.name, indent(data_str))

    # TODO implement flattened versions of packet object to allow for packet forwarding
    # naively, this can be done just by converting to a tuple of setting/parameter tuples,
    # but we would also like to preserve the type hints provided by the accepted types
    # for the settings themselves, so we can either flatten right here or else pass that
    # information along to be used later.
    #def __lrtype__(self):
    #    pass

    #def __lrflatten__(self):
    #    pass

def unwrap(s, after='|'):
    def trim(line):
        return line.split(after, 1)[-1]
    return '\n'.join(trim(line) for line in s.split('\n'))

def makePacketWrapperClass(server):
    """Make a new packet wrapper class for a particular server."""
    class CustomPacketWrapper(AsyncPacketWrapper):
        settings = MultiDict()
        _server = server
        _cache = {}
    return CustomPacketWrapper


class AsyncServerWrapper(object):
    """Represents a remote LabRAD server."""

    def __init__(self, cxn, name, pyName, ID):
        self._cxn = cxn
        self._mgr = cxn._mgr
        self.name = self._labrad_name = name
        self._py_name = pyName
        self.ID = ID
        self.settings = MultiDict()
        self._cache = {}
        self._packetWrapperClass = makePacketWrapperClass(self)
        self._refreshLock = defer.DeferredLock()

    def refresh(self):
        return self._refreshLock.run(self._refresh)

    @inlineCallbacks
    def _refresh(self):
        """Update the list of available settings for this server."""

        # get info about this server and its settings
        info = yield self._mgr.getServerInfoWithSettings(self.ID)
        self.__doc__, self.notes, settings = info
        names = [s[0] for s in settings]

        # determine what to add, update and delete to be current
        for s in settings:
            if s[0] not in self.settings:
                self._addSetting(*s)
            else:
                self._refreshSetting(*s)
        for n in self.settings:
            if n not in names:
                self._delSetting(n)

    _staticAttrs = ['settings', 'refresh', 'context', 'packet',
                    'sendMessage', 'addListener', 'removeListener']

    def _fixName(self, name):
        pyName = mangle(name)
        if pyName in self._staticAttrs:
            pyName = 'lr_' + pyName
        return pyName

    def _addSetting(self, name, ID, info):
        """Add a wrapper for a new setting for this server.

        The wrapper will be pulled from the cache and reused if
        one has already been created for a setting with this name.
        Also add this setting to the packet wrapper class.
        """
        if name in self._cache:
            setting = self._cache[name]
            setting._updateInfo(ID, info)
        else:
            pyName = self._fixName(name)
            setting = AsyncSettingWrapper(self, name, pyName, ID, info)
            self._cache[name] = setting
        self.settings[name, setting._py_name, ID] = setting
        setattr(self, setting._py_name, setting)
        # also add to the packet class
        self._packetWrapperClass._addSetting(setting)

    def _refreshSetting(self, name, ID, info):
        """Refresh the info about a particular setting.

        In particular, update the MultiDict mapping if the
        setting ID has changed.
        """
        setting = self.settings[name]
        aliasChanged = (ID != setting.ID)
        setting._updateInfo(ID, info)
        if aliasChanged:
            # update aliases if the ID has changed
            self.settings._updateAliases(name, setting._py_name, ID)
            self._packetWrapperClass._refreshSetting(setting)

    def _delSetting(self, name):
        """Remove the wrapper for a setting."""
        setting = self.settings[name]
        del self.settings[name]
        delattr(self, setting._py_name)
        # also remove from the packet class
        self._packetWrapperClass._delSetting(setting)

    def context(self):
        """Create a new context for talking to this server."""
        return self._cxn.context()

    def packet(self, **kw):
        """Create a new packet for this server."""
        return self._packetWrapperClass(**kw)

    def sendMessage(self, ID, *args, **kw):
        """Send a message to this server."""
        tag = extractKey(kw, 'tag', [])
        if len(args) == 0:
            args = None
        elif len(args) == 1:
            args = args[0]
        self._cxn._sendMessage(self.ID, [(ID, args, tag)], **kw)

    def addListener(self, listener, **kw):
        """Add a listener for messages from this server."""
        kw['source'] = self.ID
        self._cxn._addListener(listener, **kw)

    def removeListener(self, listener, **kw):
        """Remove a listener for messages from this server."""
        kw['source'] = self.ID
        self._cxn._removeListener(listener, **kw)

    def __call__(self):
        return self
        # TODO this should make a clone that can have different
        # default keyword args, and, in particular, should talk
        # to the server in a different context

    def __getitem__(self, key):
        return self.settings[key]

    def _send(self, *args, **kw):
        """Send packet to this server."""
        return self._cxn._send(self.ID, *args, **kw)

@inlineCallbacks
def getConnection(host=C.MANAGER_HOST, port=None, name="Python Client",
                  password=None, tls_mode=C.MANAGER_TLS, username=None,
                  headless=False):
    """Connect to LabRAD and return a deferred that fires the protocol object."""
    p = yield protocol.connect(host, port, tls_mode)
    yield p.authenticate(username, password, headless)
    yield p.loginClient(name)
    returnValue(p)

@inlineCallbacks
def connectAsync(host=C.MANAGER_HOST, port=None, name="Python Client",
                 password=None, tls_mode=C.MANAGER_TLS, username=None,
                 headless=False):
    """Connect to LabRAD and return a deferred that fires the client object."""
    p = yield getConnection(host, port, name, password, tls_mode=tls_mode,
                            username=username, headless=headless)
    cxn = ClientAsync(p)
    yield cxn._init()
    cxn.onDisconnect = p.onDisconnect
    returnValue(cxn)

def runAsync(func, *args, **kw):
    from twisted.internet import reactor
    @inlineCallbacks
    def runIt():
        try:
            yield func(*args, **kw)
        finally:
            reactor.stop()
    reactor.callWhenRunning(runIt)
    reactor.run()

class ClientAsync(object):
    """Adapt a LabRAD request protocol object to an asynchronous client."""

    def __init__(self, prot):
        self.servers = MultiDict()
        self._cache = {}
        self._cxn = prot
        self._mgr = manager.AsyncManager(self._cxn)
        self._refreshLock = defer.DeferredLock()

    _staticAttrs = ['servers', 'refresh', 'context']

    @inlineCallbacks
    def _init(self):
        """Refresh the cache of available servers and their settings.

        Also set up messages so that servers that connect and disconnect later
        will be automatically detected, without needing a refresh.
        """
        try:
            yield self._mgr.subscribeToNamedMessage('Server Connect', 314159265, True)
            yield self._mgr.subscribeToNamedMessage('Server Disconnect', 314159266, True)
            self._cxn.addListener(self._serverConnected, source=self._mgr.ID, ID=314159265, async=False)
            self._cxn.addListener(self._serverDisconnected, source=self._mgr.ID, ID=314159266, async=False)
            yield self.refresh()
        except Exception, e:
            print 'error!'
            print repr(e)
            raise

    @inlineCallbacks
    def _serverConnected(self, _c, data):
        """Add a wrapper when a server connects."""
        ID, name = data
        try:
            yield self._addServer(name, ID)
        except Exception, e:
            print 'Error adding server %d, "%s":' % (ID, name)
            print str(e)

    @inlineCallbacks
    def _serverDisconnected(self, _c, data):
        """Remove the wrapper when a server disconnects."""
        ID, name = data
        try:
            yield self._delServer(name)
        except Exception, e:
            print 'Error removing server %d, "%s":' % (ID, name)
            print str(e)

    def refresh(self):
        return self._refreshLock.run(self._refresh)

    @inlineCallbacks
    def _refresh(self):
        """Update the list of available LabRAD servers."""

        # get a list of the currently-available servers
        slist = yield self._mgr.getServersList()
        names = [s[0] for s in slist]

        # determine what to add, update and delete to be current
        additions = [s for s in slist if s[0] not in self.servers]
        refreshes = [n for n in self.servers if n in names]
        deletions = [n for n in self.servers if n not in names]

        actions = ([self._addServer(*s) for s in additions] +
                   [self._refreshServer(n) for n in refreshes] +
                   [self._delServer(n) for n in deletions])
        yield defer.DeferredList(actions, fireOnOneErrback=True)

    def _fixName(self, name):
        pyName = mangle(name)
        if pyName in self._staticAttrs:
            pyName = 'lr_' + pyName
        return pyName

    @inlineCallbacks
    def _addServer(self, name, ID):
        """Create a wrapper for a new server and add it to the list.

        Wrappers are cached so that if a server stops and restarts,
        the old wrapper will be reused and references to it do not
        need to be updated.
        """
        if name in self._cache:
            # pull from cache if possible
            server = self._cache[name]
        else:
            # create a new wrapper and cache it
            pyName = self._fixName(name)
            server = AsyncServerWrapper(self, name, pyName, ID)
            self._cache[name] = server
        try:
            yield server.refresh()
        except Exception, e:
            print 'Error while refreshing server "%s":' % name
            print repr(e)
        else:
            self.servers[name, server._py_name, ID] = server
            setattr(self, server._py_name, server)

    @inlineCallbacks
    def _refreshServer(self, name):
        """Trigger a refresh on a server wrapper."""
        server = self.servers[name]
        try:
            yield server.refresh()
        except Exception, e:
            print 'Error while refreshing server "%s":' % name
            print repr(e)
            yield self._delServer(name)

    def _delServer(self, name):
        """Remove a server wrapper.

        The wrapper is kept in the cache, to be reused if this
        server ever comes back.
        """
        if name in self.servers:
            server = self.servers[name]
            del self.servers[name]
            delattr(self, server._py_name)
        return defer.succeed(None)

    def context(self):
        """Create a new communication context for this connection."""
        return self._cxn.context()

    def __getitem__(self, key):
        return self.servers[key]

    def _send(self, target, records, *args, **kw):
        """Send a packet over this connection."""
        return self._cxn.sendRequest(target, records, *args, **kw)

    def _sendMessage(self, target, records, *args, **kw):
        """Send a message over this connection."""
        return self._cxn.sendMessage(target, records, *args, **kw)

    def disconnect(self):
        self._cxn.disconnect()

    @property
    def _addListener(self):
        return self._cxn.addListener

    @property
    def _removeListener(self):
        return self._cxn.removeListener

@inlineCallbacks
def wrapAsync(cls, *args, **kw):
    obj = cls(*args, **kw)
    yield obj.refresh()
    returnValue(obj)



# Copyright (C) 2007  Matthew Neeley
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
labrad.node

Provides an interface to manage multiple labrad servers running on a particular
computer. The node runs each server in a separate process, so that they can not
interfere with each other. Information such as the manager host, port and
password are passed to the child process in environment variables or via
command line arguments. The startup process for each child server is controlled
by an associated .ini file.

The node name is read either from the LABRADNODE environment variable,
or failing that the system's hostname.

The node running on a particular computer is configured via the registry, at
the path ['', 'Nodes', node_name]. Defaults values can be configured to be
used whenever a new node connects, at ['', 'Nodes', '__default__']. The
configuration keys are:

  directories (*s): file directories to be scanned for servers. Any subdirs
                    that contain a file called '.nodeignore' will be skipped,
                    along with all their subdirs.
  extensions (*s): what files to look at, e.g. ['.ini', '.py', '.exe']
  autostart (*s): list of servers to be autostarted

Changes required on the server:

See "launchable-server.ini" for a description of how to have a server appear in
the node. See labrad-servers/Qubit Server/ for an example of a java server.

Another option is to have this configuration in the file itself, set off by
### BEGIN NODE INFO and ### END NODE INFO. See servers/gpibMockDeviceServer.py
in this repository for an example.

***N.B.*** The name given in the server's [info] section MUST MATCH the name
given as a class variable in the server class itself. That is, note that the
name string "GPIB Mock Device Server" appears _twice_. If this is not the case,
then while the node will still start the server, it will not register that the
server has been started and you will have to drop the server connection manually
to restart it.

Bonus: The node will also open up compiled executable files to look for
this information. This is how it works for the Direct Ethernet server, for
example.

The node can log to a file or via the syslog fascility.  Syslog can be
configured to filter messages generated by the node and its servers.
For rsyslogd on ubuntu, create the following file:

/etc/rsyslog.d/71-labrad.conf
:syslogtag,contains,"labrad"			/var/log/labrad.log;RSYSLOG_TraditionalFileFormat
"""

from __future__ import with_statement

from ConfigParser import SafeConfigParser
from datetime import datetime
import logging
import logging.handlers
import os
import shlex
import socket
import StringIO
import sys
import zipfile

from twisted.application.service import MultiService
from twisted.application.internet import TCPClient
from twisted.internet import defer, reactor
from twisted.internet.defer import inlineCallbacks, returnValue
from twisted.internet.protocol import ProcessProtocol
from twisted.internet.error import ProcessDone, ProcessTerminated
from twisted.python import usage
from twisted.python.runtime import platformType

import labrad
from labrad import auth, protocol, util, types as T, constants as C
from labrad.server import LabradServer, setting
import labrad.support
from labrad.util import dispatcher, findEnvironmentVars, interpEnvironmentVars


LOG_LENGTH = 1000 # maximum number of lines of stdout to keep per server


class ServerProcess(ProcessProtocol):
    """A class to represent a running server instance."""

    timeout = 20
    shutdownTimeout = 5

    def __init__(self, env):
        self.env = os.environ.copy()
        self.env.update(env, DIR=self.path, FILE=self.filename)
        cls = self.__class__
        self.name = interpEnvironmentVars(cls.instancename, self.env)
        self.args = shlex.split(self.cmdline)
        self.args = [interpEnvironmentVars(a, self.env) for a in self.args]
        self.starting = False
        self.started = False
        self.stopping = False
        self.output = []
        self._lock = defer.DeferredLock()
        logname = 'labrad.' + labrad.support.mangle(self.name)
        self.logger = logging.getLogger(logname)

    @property
    def status(self):
        if self.starting:
            return 'STARTING'
        elif self.started:
            return 'STARTED'
        elif self.stopping:
            return 'STOPPING'
        else:
            return 'STOPPED'

    def start(self):
        """Start this server instance."""
        return self._lock.run(self._start)

    def stop(self):
        """Stop this server instance."""
        return self._lock.run(self._stop)

    def restart(self):
        """Restart this server instance."""
        return self._lock.run(self._restart)

    @inlineCallbacks
    def _start(self):
        if self.started:
            return
        print "starting '%s'..." % self.name
        print "path:", self.path
        print "args:", self.args
        self.starting = True
        self.startup = defer.Deferred()
        dispatcher.connect(self.serverConnected, 'serverConnected')
        self.emitMessage('server_starting')
        # This looks crazy.  On Unix, spawnProcess calls fork() and then
        # execvpe which expects an executable name and a series of arguments,
        # and tries to resolve the executable using PATH.  On windows,
        # it calls createProcess, which expects a command string and an
        # optional executable.  If the executable argument is present,
        # it must be an absolute path including .exe suffix.  If executable
        # is not present, the first part of the command string is used,
        # and invokes the %PATH% and extension search.
        #
        # Twisted spawnProcess only supports posix (including Mac) and
        # win32, so we just abandon anything else.
        # Relevant code is in twisted/internet/process.py:_BaseProcess and
        # twisted/internet/dumbwin32proc.py:Process. The twisted docs for
        # spawnProcess say that the full executable path is required, but
        # this is not in fact the case. See:
        # http://twistedmatrix.com/documents/current/api/twisted.internet.interfaces.IReactorProcess.spawnProcess.html
        if platformType == 'posix':
            executable = self.args[0]
        elif platformType == 'win32':
            executable = None
        else:
            raise RuntimeError("Unsupported platform %s" % platformType)

        self.proc = reactor.spawnProcess(self, executable, self.args,
                                         env=self.env, path=self.path)
        timeoutCall = reactor.callLater(self.timeout, self.kill)
        try:
            yield self.startup
            self.emitMessage('server_started')
        except:
            self.emitMessage('server_stopped')
            raise
        finally:
            self.starting = False
            if timeoutCall.active():
                timeoutCall.cancel()
            try:
                dispatcher.disconnect(self.serverConnected, 'serverConnected')
            except Exception, e:
                print 'Error while disconnecting signal:', e

    @inlineCallbacks
    def _stop(self):
        if not self.started:
            return
        print "stopping '%s'..." % self.name
        self.stopping = True
        self.shutdown = defer.Deferred()
        self.emitMessage('server_stopping')
        # hack: marker to tell the kill func that it worked
        finished = [False]
        self.kill(finished)
        yield self.shutdown
        finished[0] = True
        self.stopping = False
        self.emitMessage('server_stopped')

    @inlineCallbacks
    def _restart(self):
        yield self._stop()
        yield self._start()

    def emitMessage(self, msg):
        """Emit a message to other parts of this application."""
        dispatcher.send(msg,
                        sender=self,
                        server=self.__class__.name,
                        instance=self.name)

    def serverConnected(self, ID, name):
        """Called when a server connects to LabRAD.

        If the name matches our name, we'll assume this server
        started successfully.  This may not be the case (e.g. if
        two nodes are trying to start the same server simultaneously),
        but there's no way to find out from LabRAD which node
        a given server is running on, so this will have to do.
        """
        if name == self.name:
            self.ID = ID
            if self.starting:
                self.started = True
                self.startup.callback(self)

    def processEnded(self, reason):
        """Called when the server process ends.

        We check to see the reason why this process failed, and then
        call the appropriate deferred, depending on the current state.
        """
        if isinstance(reason.value, ProcessDone):
            print "'%s': process closed cleanly." % self.name
        elif isinstance(reason.value, ProcessTerminated):
            print "'%s': process terminated: %s" % (self.name, reason.value)
        else:
            print "'%s': process ended: %s" % (self.name, reason)
        self.started = False
        if self.starting:
            err = T.Error('Startup failed.', payload=self.output)
            self.startup.errback(err)
        elif self.stopping:
            self.shutdown.callback(None)
        else:
            # looks like this thing died on its own
            self.emitMessage('server_stopped')

    @inlineCallbacks
    def kill(self, finished=None):
        """Kill the server process."""
        if not self.started:
            return
        try:
            servers = self.client.servers
            if hasattr(self, 'shutdownMode') and self.name in servers:
                mode, ID = self.shutdownMode
                if mode == 'message':
                    # try to shutdown by sending a message
                    servers[self.name].sendMessage(ID)
                elif mode == 'setting':
                    # try to shutdown by calling a setting
                    try:
                        yield servers[self.name][ID]()
                    except:
                        pass
                yield util.wakeupCall(self.shutdownTimeout)

            # hack to let us know that we did indeed finish killing the server
            if finished is not None:
                if finished[0]:
                    return

            # if we're not dead yet, kill with a vengeance
            if self.started:
                self.proc.signalProcess('KILL')
        except Exception:
            logging.error('Error while trying to kill server process for "%s":' % self.name,
                          exc_info=True)

    def outReceived(self, data):
        """Called when the server prints to stdout."""
        self.output.append((datetime.now(), data))
        self.output = self.output[-LOG_LENGTH:]
        self.logger.info(data.strip())

    def errReceived(self, data):
        """Called when the server prints to stderr."""
        self.output.append((datetime.now(), data))
        self.output = self.output[-LOG_LENGTH:]
        self.logger.warning(data.strip())

    def clearOutput(self):
        """Clear the log of stdout."""
        self.output = []


def findConfigBlock(path, filename):
    """Find a Node configuration block embedded in a file."""
    # markers to delimit node info block
    BEGIN = "### BEGIN NODE INFO"
    END = "### END NODE INFO"
    with open(os.path.join(path, filename), 'rb') as file:
        foundBeginning = False
        lines = []
        for line in file:
            if line.upper().strip().startswith(BEGIN):
                foundBeginning = True
            elif line.upper().strip().startswith(END):
                break
            elif foundBeginning:
                line = line.replace('\r', '')
                line = line.replace('\n', '')
                lines.append(line)
        return '\n'.join(lines) if lines else None


def createGenericServerCls(path, filename, conf):
    """Create a ServerProcess class representing a generic server.

    Options for this server are passed in as a string in standard
    .ini format.  We use a string rather than a file to allow this
    configuration to be extracted from a larger file if necessary.
    """
    class cls(ServerProcess):
        pass

    scp = SafeConfigParser()
    scp.readfp(StringIO.StringIO(conf))

    # general information
    cls.name = scp.get('info', 'name', raw=True)
    cls.__doc__ = scp.get('info', 'description', raw=True)
    if scp.has_option('info', 'version'):
        cls.version = scp.get('info', 'version', raw=True)
    else:
        cls.version = '0.0'
    try:
        cls.instancename = scp.get('info', 'instancename', raw=True)
    except:
        cls.instancename = cls.name
    cls.environVars = findEnvironmentVars(cls.instancename)
    cls.isLocal = len(cls.environVars) > 0

    # startup
    platform_cmdline_option = 'cmdline_{}'.format(sys.platform)
    if scp.has_option('startup', platform_cmdline_option):
        # use platform-specific command line
        cls.cmdline = scp.get('startup', platform_cmdline_option, raw=True)
    else:
        # use generic command line
        cls.cmdline = scp.get('startup', 'cmdline', raw=True)
    cls.path = path
    cls.filename = filename
    try:
        cls.timeout = float(scp.getint('startup', 'timeout'))
    except:
        pass

    # shutdown
    if scp.has_option('shutdown', 'message'):
        cls.shutdownMode = 'message', int(scp.get('shutdown', 'message', raw=True))
    elif scp.has_option('shutdown', 'setting'):
        cls.shutdownMode = 'setting', scp.get('shutdown', 'setting', raw=True)
    try:
        cls.shutdownTimeout = float(scp.getint('shutdown', 'timeout'))
    except:
        pass

    return cls


def version_tuple(version):
    """Get a tuple from a version string that can be used for comparison.

    Version strings are typically of the form A.B.C-X where A, B and C
    are numbers, and X is extra text denoting dev status (e.g. alpha or beta).
    Given this structure, we cannot just use string comparison to get the order
    of versions; instead we parse the version into a tuple

    ((int(A), int(B), int(C)), version)

    If we cannot parse the numeric part, we just use the empty tuple for the
    first entry, and for such tuples the comparison will just fall back to
    alphabetic comparison on the full versions string.
    """
    numstr, _, _extra = version.partition('-')
    try:
        nums = tuple(int(n) for n in numstr.split('.'))
    except Exception:
        nums = ()
    return (nums, version)


class Node(object):
    """Parent class that keeps the node running.

    If the manager is stopped or we lose the network connection,
    this service attempts to restart it so that we will come
    back online when the manager is back up.
    """
    reconnectDelay = 10

    def __init__(self, nodename, host, port, username, password,
                 tls_mode=C.MANAGER_TLS):
        self.nodename = nodename
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.tls_mode = tls_mode

    @inlineCallbacks
    def run(self):
        """Run the node in a loop, reconnecting after connection loss."""
        log = logging.getLogger('labrad.node')
        while True:
            print 'Connecting to {}:{}...'.format(self.host, self.port)
            try:
                p = yield protocol.connect(self.host, self.port, self.tls_mode)
                yield p.authenticate(self.username, self.password)
                node = NodeServer(self.nodename, self.host, self.port,
                                  p.credential)
                yield node.startup(p)
            except Exception:
                log.error('Node failed to start', exc_info=True)
            else:
                try:
                    yield node.onShutdown()
                except Exception:
                    log.error('Error during node shutdown', exc_info=True)

            ## hack: manually clear the internal message dispatcher
            dispatcher.connections.clear()
            dispatcher.senders.clear()
            dispatcher._boundMethods.clear()

            yield util.wakeupCall(0)
            print 'Will try to reconnect in {} seconds...'.format(self.reconnectDelay)
            yield util.wakeupCall(self.reconnectDelay)


class NodeConfig(object):
    """Load configuration from the registry and monitor it for changes.

    Attributes:
        dirs (list(string)): a list of directories that will be searched for
            runnable servers.
        exts (list(string)): a list of file extensions that will be included
            when searching for servers.
        autostart (list(string)): a list of servers that will be automatically
            started when the node launches.
    """

    @classmethod
    @inlineCallbacks
    def create(cls, parent):
        """Loads node configuration from the registry."""
        instance = cls(parent)
        yield instance._init()
        returnValue(instance)

    def __init__(self, parent):
        self.parent = parent
        self.nodename = parent.nodename
        cxn = parent.client
        self._cxn = cxn
        self._reg = cxn.registry
        self._ctx = cxn.context()

    @inlineCallbacks
    def _init(self):
        """Initialize by loading from the registry.

        Copy from the default directory, creating it if necessary.
        Also, set up messages to monitor the config directory for
        changes.
        """
        p = self._packet()
        p.cd(['', 'Nodes'], True)
        p.dir()
        ans = yield p.send()
        dirs, keys = ans.dir

        # load defaults (creating them if necessary)
        create = '__default__' not in dirs
        defaults = ([], ['.ini', '.py'], [])
        defaults = yield self._load('__default__', create, defaults)

        # load this node (creating config if necessary)
        create = self.nodename not in dirs
        config = yield self._load(self.nodename, create, defaults)
        self._update(config, False)

        # setup messages when registry changes
        self._reg.addListener(self._handleMessage, context=self._ctx)
        p = self._packet()
        p.notify_on_change(2345, True)
        yield p.send()

    def _packet(self):
        """Create a packet to the registry server in our context."""
        return self._reg.packet(context=self._ctx)

    def _update(self, config, triggerRefresh=True):
        """Update instance variables from loaded config."""
        self.dirs, self.extensions, self.autostart = config
        print 'config updated: dirs={}, extensions={}, autostart={}'.format(
                self.dirs, self.extensions, self.autostart)
        if triggerRefresh:
            self.parent.refreshServers()

    @inlineCallbacks
    def _load(self, nodename=None, create=False, defaults=None):
        """Load the current configuration out of the registry."""
        p = self._packet()
        if nodename is not None:
            p.cd(['', 'Nodes', nodename], True)
        if create:
            p.set('directories', defaults[0])
            p.set('extensions', defaults[1])
            p.set('autostart', defaults[2])
        p.get('directories', '*s', key='dirs')
        p.get('extensions', '*s', key='exts')
        p.get('autostart', '*s', True, [], key='autostart')
        ans = yield p.send()
        def remove_empties(strs):
            return [s for s in strs if s]
        dirs = remove_empties(ans.dirs)
        exts = remove_empties(ans.exts)
        autostart = sorted(remove_empties(ans.autostart))
        returnValue((dirs, exts, autostart))

    def _save(self):
        """Save the current configuration to the registry."""
        p = self._packet()
        p.set('directories', self.dirs)
        p.set('extensions', self.extensions)
        return p.send()

    @inlineCallbacks
    def _handleMessage(self, c, msg):
        """Reload when we get a message from the registry."""
        try:
            config = yield self._load()
            self._update(config)
        except Exception:
            logging.error('Error in _handleMessage', exc_info=True)

    @inlineCallbacks
    def update_autostart(self, autostart):
        """Update the list of servers to be autostarted.

        Args:
            autostart (list(string)): New list of autostart server names. This
                will completely replace the current list.
        """
        p = self._packet()
        p.cd(['', 'Nodes', self.nodename])
        p.set('autostart', sorted(autostart))
        yield p.send()


class NodeServer(LabradServer):
    """Start and stop LabRAD servers remotely.

    The node server allows you to control and
    monitor servers running on a remote machine.
    """

    name = 'node %LABRADNODE%'

    def __init__(self, nodename, host, port, credential):
        LabradServer.__init__(self)
        self.nodename = nodename
        self.name = 'node %s' % nodename
        self.host = host
        self.port = port
        self.credential = credential
        self.servers = {}
        self.instances = {}
        self.starters = {}
        self.runners = {}
        self.stoppers = {}
        self.initMessages(True)

    @inlineCallbacks
    def initServer(self):
        """Initialize this server."""
        self.config = yield NodeConfig.create(self)
        self.refreshServers()
        self.autostart(None)

    def stopServer(self):
        """Stop this node by killing all subprocesses."""
        self.initMessages(False)
        stoppages = [srv.stop() for srv in self.runners.values()]
        return defer.DeferredList(stoppages)


    # message handling

    def initMessages(self, connect=True):
        """Set up message dispatching."""
        attr = 'connect' if connect else 'disconnect'
        method = getattr(dispatcher, attr)
        def f(receiver, signal):
            try:
                method(receiver, signal)
            except dispatcher.DispatcherError as e:
                msg = 'Error while setting up message dispatching. receiver={0}, method={1}, signal={2}.'
                print msg.format(receiver, attr, signal), e
        # set up messages to be relayed out over LabRAD
        messages = ['server_starting', 'server_started',
                    'server_stopping', 'server_stopped',
                    'status']
        for message in messages:
            f(self._relayMessage, message)
        # set up message handlers for subprocess events
        f(self.subprocessStarting, 'server_starting')
        f(self.subprocessStarted, 'server_started')
        f(self.subprocessStopping, 'server_stopping')
        f(self.subprocessStopped, 'server_stopped')

    def _relayMessage(self, signal, sender, **kw):
        """Send messages out to LabRAD."""
        kw['node'] = self.name
        mgr = self.client.manager
        mgr.send_named_message('node.' + signal, tuple(kw.items()))

    def serverConnected(self, ID, name):
        """Called when a server connects to LabRAD."""
        dispatcher.send('serverConnected', ID=ID, name=name)

    def subprocessStarting(self, sender):
        """Called when a subprocess begins connecting."""
        self.starters[sender.name] = sender

    def subprocessStarted(self, sender):
        """Called when a subprocess successfully connects."""
        if sender.name in self.starters:
            del self.starters[sender.name]
        self.runners[sender.name] = sender

    def subprocessStopping(self, sender):
        """Called when a subprocess successfully disconnects."""
        if sender.name in self.runners:
            del self.runners[sender.name]
        self.stoppers[sender.name] = sender

    def subprocessStopped(self, sender):
        """Called when a subprocess successfully disconnects."""
        if sender.name in self.runners:
            del self.runners[sender.name]
        if sender.name in self.stoppers:
            del self.stoppers[sender.name]


    # status information

    def status(self):
        """Get information about all servers on this node."""
        def serverInfo(cls):
            instances = [n for n, s in self.runners.items()
                                    if s.__class__.name == cls.name]
            return (cls.name, cls.__doc__ or '', cls.version,
                    cls.instancename, cls.environVars, instances)
        return [serverInfo(item[1])
                for item in sorted(self.servers.items())]


    # server refresh

    def refreshServers(self):
        """Refresh the list of available servers."""
        # configs is a nested map from name to version to list of classes.
        #
        # This allows us to deal with cases where there are many definitions
        # for different server versions, and possibly also redundant defitions
        # for the same version.
        configs = {}

        # look for .ini files
        for dirname in self.config.dirs:
            for path, dirs, files in os.walk(dirname):
                if '.nodeignore' in files:
                    del dirs[:] # clear dirs list so we don't visit subdirs
                    continue
                for f in files:
                    try:
                        _, ext = os.path.splitext(f)
                        if ext.lower() not in self.config.extensions:
                            continue
                        if ext.lower() == '.ini':
                            with open(os.path.join(path, f)) as file:
                                conf = file.read()
                        elif ext.lower() in ['.war', '.jar']:
                            zf = zipfile.ZipFile(os.path.join(path, f))
                            found = False
                            for info in zf.infolist():
                                if 'node.ini' in info.filename:
                                    found = True
                                    member = zf.open(info.filename)
                                    conf = member.read()
                                    member.close()
                                    break
                            zf.close()
                        else:
                            conf = findConfigBlock(path, f)
                            if conf is None:
                                continue
                        s = createGenericServerCls(path, f, conf)
                        s.client = self.client
                        if s.name not in configs:
                            configs[s.name] = {}
                        versions = configs.setdefault(s.name, {})
                        classes = versions.setdefault(s.version, [])
                        classes.append(s)
                    except Exception:
                        fname = os.path.join(path, f)
                        logging.error('Error while loading config file "%s":' % fname,
                                  exc_info=True)

        servers_dict = {}
        for versions in configs.values():
            for servers in versions.values():
                if len(servers) > 1:
                    conflicting_files = [s.filename for s in servers]
                    s = servers[0]
                    logging.warning(
                        'Found redundant server configs with same name and '
                        'version; will use {}. name={}, version={}, '
                        'conflicting_files={}'
                        .format(s.filename, s.name, s.version,
                                conflicting_files))

            servers = [ss[0] for ss in versions.values()]
            servers.sort(key=lambda s: version_tuple(s.version))
            if len(servers) > 1:
                # modify server name for all but the latest version
                for s in servers[:-1]:
                    s.name = '{}-{}'.format(s.name, s.version)

            for s in servers:
                servers_dict[s.name] = s
        self.servers = servers_dict
        # send a message with the current server list
        dispatcher.send('status', servers=self.status())


    # LabRAD settings

    @setting(1, name='s', environ='*(ss)', returns='s')
    def start(self, c, name, environ={}):
        """Start an instance of a server."""
        if name not in self.servers:
            raise Exception("Unknown server: '%s'." % name)
        environ = dict(environ) # copy context environment
        environ.update(LABRADNODE=self.nodename,
                       LABRADHOST=self.host,
                       LABRADPORT=str(self.port),
                       PYTHON=sys.executable)
        if isinstance(self.credential, auth.Password):
            environ.update(LABRADUSER=self.credential.username,
                           LABRADPASSWORD=self.credential.password)
        srv = self.servers[name](environ)
        # TODO check whether an instance with this name already exists
        self.instances[name] = srv
        yield srv.start()
        returnValue(srv.name)

    @setting(2, name='s', returns='s')
    def stop(self, c, name):
        """Stop a running server instance."""
        if name not in self.runners:
            raise Exception("'%s' is not running." % name)
        srv = self.runners[name]
        yield srv.stop()
        returnValue(srv.name)

    @setting(3, name='s', returns='s')
    def restart(self, c, name):
        """Restart a running server instance."""
        if name not in self.runners:
            raise Exception("'%s' is not running." % name)
        srv = self.runners[name]
        yield srv.restart()
        returnValue(srv.name)

    @setting(10, returns='*s')
    def available_servers(self, c):
        """Get a list of available servers."""
        return sorted(self.servers.keys())

    @setting(11, returns='*(ss)')
    def running_servers(self, c):
        """Get a list of running server instances.

        Returns a list of tuples of server name and instance name.
        """
        return sorted((s.__class__.name, n) for n, s in self.runners.items())

    @setting(12, returns='*s')
    def local_servers(self, c):
        """Get a list of local servers."""
        return sorted(n for n, s in self.servers.items() if s.isLocal)

    @setting(13, returns='')
    def refresh_servers(self, c):
        """Refresh the list of available servers."""
        yield self.refreshServers()

    @setting(14, 'status',
             returns='*(s{name} s{desc} s{ver} s{instname} *s{vars} *s{running})')
    def get_status(self, c):
        """Get information about all servers on this node."""
        return self.status()

    @setting(100, name='s', returns='*(ts)')
    def server_output(self, c, name):
        """Get output from a server's stdout."""
        if name not in self.runners:
            raise Exception("'%s' is not running." % name)
        return self.runners[name].output

    @setting(101, name='s', returns='')
    def clear_output(self, c, name):
        """Clear the stdout buffer of a server."""
        if name not in self.runners:
            raise Exception("'%s' is not running." % name)
        self.runners[name].clearOutput()

    @setting(102, name='s', returns='s')
    def server_version(self, c, name):
        """Get version information for a server."""
        if name not in self.servers:
            raise Exception("'%s' not found." % name)
        return self.servers[name].version

    @setting(103, name='s', enable='b', returns='')
    def stream_output(self, c, name, enable):
        """Enable or disable server output messages.

        This allows you to receive messages whenever a server
        outputs something on its stdout, effectively giving a
        remote view of the server's console window.
        """
        pass

    @setting(200, returns='')
    def autostart(self, c):
        """Start all servers from the configured autostart list.

        Any servers that are already running will be left as is, while those
        that are not yet running will be started. Autostart is triggered when
        the node first starts up, but can be invoked manually at any time
        thereafter.
        """
        running = set(s.__class__.name for s in self.runners.values())
        to_start = [name for name in self.config.autostart
                         if name not in running]
        deferreds = [(name, self.start(c, name)) for name in to_start]
        for name, deferred in deferreds:
            try:
                yield deferred
            except Exception:
                logging.error('Failed to autostart "%s"', name, exc_info=True)

    @setting(201, returns='*s')
    def autostart_list(self, c):
        """Get the list of servers that are configured to be autostarted."""
        return self.config.autostart

    @setting(202, name='s', returns='')
    def autostart_add(self, c, name):
        """Add a server to the autostart list."""
        if name not in self.servers:
            raise Exception("Unknown server: '%s'." % name)
        autostart = set(self.config.autostart)
        autostart.add(name)
        yield self.config.update_autostart(sorted(autostart))

    @setting(203, name='s', returns='')
    def autostart_remove(self, c, name):
        """Remove a server from the autostart list."""
        autostart = set(self.config.autostart)
        try:
            autostart.remove(name)
        except KeyError:
            pass
        yield self.config.update_autostart(sorted(autostart))

    @setting(1000, returns='*(ss)')
    def node_version(self, c):
        """Return a list of key-value tuples containing info about this node."""
        info = {
            'hostname': socket.gethostname(),
            'nodename': self.nodename,
            'python version': sys.version,
            'labrad version': labrad.__version__,
            'labrad revision': labrad.__revision__,
            'labrad date': labrad.__date__,
            }
        return list(info.items())


class NodeOptions(usage.Options):
    optParameters = [
            ['name', 'n', util.getNodeName(), 'Node name.'],
            ['host', 'h', C.MANAGER_HOST, 'Manager location.'],
            ['port', 'p', C.MANAGER_PORT, 'Manager port.'],
            ['username', 'u', None, 'Login username.'],
            ['password', 'w', None, 'Login password.'],
            ['tls', '', C.MANAGER_TLS,
             'TLS mode for connecting to manager (on/starttls/off)'],
            ['logfile', 'l', None, 'Enable logging to a file'],
            ['syslog_socket', 'x', None,
             'Override default syslog socket. Absolute path or host[:port]']]
    optFlags = [['syslog', 's', 'Enable syslog'],
                ['verbose', 'v', 'Enable debug output']]


def makeService(options):
    """Construct a TCPServer from a LabRAD node."""
    name = options['name']
    host = options['host']
    port = int(options['port'])
    username = options['username']
    password = options['password']
    tls_mode = C.check_tls_mode(options['tls'])
    return Node(name, host, port, username, password, tls_mode)


def setup_logging(options):
    logging.basicConfig()
    node_log = logging.getLogger('labrad')
    if options['syslog']:
        # We need to find the path to the system log socket, which varies by
        # platform. Linux and OS/X defaults are listed below. On windows the
        # only option is UDP logging, but since UDP is connectionless there is
        # no way to tell if there is actually a syslog daemon listening.
        # https://docs.python.org/2/library/logging.handlers.html#sysloghandler
        if options['syslog_socket']:
            if '/' in options['syslog_socket']:
                address = options['syslog_socket']
            else:
                host, _, port = options['syslog_socket'].partition(':')
                if port == '':
                    address = (host, 514)
                else:
                    address = (host, int(port))
        elif sys.platform.startswith('linux'):
            address = '/dev/log'
        elif sys.platform.startswith('darwin'):
            address = '/var/run/syslog'
        else:
            node_log.critical(
                    'Syslog specified, but default socket not known for '
                    'platform {}. Use -s option'.format(sys.platform))
            sys.exit(1)
        syslog_handler = logging.handlers.SysLogHandler(address=address)
        syslog_handler.setFormatter(logging.Formatter('%(name)s: %(message)s'))
        node_log.addHandler(syslog_handler)
    if options['logfile']:
        file_handler = logging.handlers.RotatingFileHandler(
                options['logfile'], maxBytes=800000, backupCount=5)
        formatter = logging.Formatter('%(asctime)s - %(name)s: %(message)s',
                                      datefmt='%Y-%m-%d %H:%M:%S')
        file_handler.setFormatter(formatter)
        node_log.addHandler(file_handler)
    if options['verbose']:
        node_log.setLevel(logging.DEBUG)
    else:
        node_log.setLevel(logging.INFO)

def main():
    config = NodeOptions()
    config.parseOptions()
    setup_logging(config)
    logging.getLogger('labrad.node').info('Starting')
    service = makeService(config)
    service.run()
    reactor.run()

if __name__ == '__main__':
    main()
